import heapq
import itertools
from collections.abc import MutableSequence
from typing import (
    Union,
    Iterable,
    Iterator,
    Optional,
    Generator,
    Dict,
    Callable,
    List,
    TYPE_CHECKING,
)

from .mixins import AllMixins
from ..document import Document
from ...helper import typename
from ...proto import jina_pb2

if TYPE_CHECKING:
    from .types import DocumentArraySourceType

try:
    # when protobuf using Cpp backend
    from google.protobuf.pyext._message import (
        RepeatedCompositeContainer as RepeatedContainer,
    )
except:
    # when protobuf using Python backend
    from google.protobuf.internal.containers import (
        RepeatedCompositeFieldContainer as RepeatedContainer,
    )

__all__ = ['DocumentArray']


class DocumentArray(
    AllMixins,
    MutableSequence,
):
    """
    :class:`DocumentArray` is a mutable sequence of :class:`Document`.
    It gives an efficient view of a list of Document. One can iterate over it like
    a generator but ALSO modify it, count it, get item, or union two 'DocumentArray's using the '+' and '+=' operators.

    It is supposed to act as a view containing a pointer to a `RepeatedContainer` of `DocumentProto` while offering `Document` Jina native types
    when getting items or iterating over it

    :param docs: the document array to construct from. One can also give `DocumentArrayProto` directly, then depending on the ``copy``,
                it builds a view or a copy from it. It also can accept a List
    """

    def __init__(self, docs: Optional['DocumentArraySourceType'] = None):
        super().__init__()
        self._pb_body = []
        if docs is not None:
            if isinstance(docs, jina_pb2.DocumentArrayProto):
                # This would happen when loading from file or memmap
                self._pb_body = docs.docs
            elif isinstance(docs, RepeatedContainer):
                # This would happen when `doc.matches` or `doc.chunks`
                self._pb_body = docs
            elif isinstance(docs, DocumentArray):
                # This would happen in the client
                self._pb_body = docs._pb_body
            else:
                if isinstance(docs, Document):
                    # single Document
                    docs = [docs]

                from .memmap import DocumentArrayMemmap

                if isinstance(
                    docs, (list, tuple, Generator, DocumentArrayMemmap, itertools.chain)
                ):
                    # This would happen in the client
                    for doc in docs:
                        if isinstance(doc, Document):
                            self._pb_body.append(doc.proto)
                        elif isinstance(doc, jina_pb2.DocumentProto):
                            self._pb_body.append(doc)
                        else:
                            raise ValueError(f'Unexpected element in an input list')
                else:
                    raise ValueError(
                        f'DocumentArray got an unexpected input {type(docs)}'
                    )
        self._id_to_index = None

    @property
    def _index_map(self) -> Dict:
        """Return the `_id_to_index` map

        :return: a Python dict.
        """
        if not self._id_to_index:
            self._update_id_to_index_map()
        return self._id_to_index

    def _update_id_to_index_map(self):
        """Update the id_to_index map by enumerating all Documents in self._pb_body.

        Very costy! Only use this function when self._pb_body is dramtically changed.
        """

        self._id_to_index = {
            d.id: i for i, d in enumerate(self._pb_body)
        }  # type: Dict[str, int]

    def insert(self, index: int, doc: 'Document') -> None:
        """Insert `doc` at `index`.

        :param index: Position of the insertion.
        :param doc: The doc needs to be inserted.
        """
        self._pb_body.insert(index, doc.proto)
        if self._id_to_index:
            self._id_to_index[doc.id] = index

    def __setitem__(self, key, value: 'Document'):
        if isinstance(key, int):
            self[key].CopyFrom(value)
            if self._id_to_index:
                self._id_to_index[value.id] = key
        elif isinstance(key, str):
            self[self._index_map[key]].CopyFrom(value)
        else:
            raise IndexError(f'do not support this index {key}')

    def __delitem__(self, index: Union[int, str, slice]):
        if isinstance(index, int):
            del self._pb_body[index]
        elif isinstance(index, str):
            del self[self._index_map[index]]
            self._index_map.pop(index)
        elif isinstance(index, slice):
            del self._pb_body[index]
        else:
            raise IndexError(
                f'do not support this index type {typename(index)}: {index}'
            )

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and type(self._pb_body) is type(other._pb_body)
            and self._pb_body == other._pb_body
        )

    def __len__(self):
        return len(self._pb_body)

    def __iter__(self) -> Iterator['Document']:
        for d in self._pb_body:
            yield Document(d)

    def __contains__(self, item: str):
        return item in self._index_map

    def __getitem__(self, item: Union[int, str, slice, List]):
        if isinstance(item, int):
            return Document(self._pb_body[item])
        elif isinstance(item, str):
            return self[self._index_map[item]]
        elif isinstance(item, slice):
            return DocumentArray(self._pb_body[item])
        elif isinstance(item, list):
            return DocumentArray(self._pb_body[t] for t in item)
        else:
            raise IndexError(f'do not support this index type {typename(item)}: {item}')

    def append(self, doc: 'Document'):
        """
        Append `doc` in :class:`DocumentArray`.

        :param doc: The doc needs to be appended.
        """
        if self._id_to_index:
            self._id_to_index[doc.id] = len(self._pb_body)
        self._pb_body.append(doc.proto)

    def extend(self, docs: Iterable['Document']) -> None:
        """
        Extend the :class:`DocumentArray` by appending all the items from the iterable.

        :param docs: the iterable of Documents to extend this array with
        """
        if not docs:
            return

        for doc in docs:
            self.append(doc)

    def clear(self):
        """Clear the data of :class:`DocumentArray`"""
        del self._pb_body[:]
        if self._id_to_index:
            self._id_to_index.clear()

    def reverse(self):
        """In-place reverse the sequence."""
        size = len(self._pb_body)
        hi_idx = size - 1
        for i in range(int(size / 2)):
            tmp = jina_pb2.DocumentProto()
            tmp.CopyFrom(self._pb_body[hi_idx])
            self._pb_body[hi_idx].CopyFrom(self._pb_body[i])
            self._pb_body[i].CopyFrom(tmp)
            hi_idx -= 1
        self._update_id_to_index_map()

    def sort(
        self,
        key: Callable,
        top_k: Optional[int] = None,
        reverse: bool = False,
    ):
        """
        Sort the items of the :class:`DocumentArray` in place.

        :param key: key callable to sort based upon
        :param top_k: make sure that the first `topk` elements are correctly sorted rather than
            sorting the entire list
        :param reverse: reverse=True will sort the list in descending order. Default is False
        """

        def overriden_key(proto):
            # Function to override the `proto` and wrap it around a `Document` to enable sorting via
            # `Document-like` interface
            d = Document(proto)
            return key(d)

        # Logic here: `overriden_key` is offered to allow the user sort via pythonic `Document` syntax. However,
        # maybe there may be cases where this won't work and the user may enter `proto-like` interface. To make
        # sure (quite fragile) the `sort` will work seamlessly, it tries to apply `key` to the first element and
        # see if it works. If it works it can sort with `proto` interface, otherwise use `Document` interface one.
        # (Very often the 2 interfaces are both the same and valid, so proto will have less overhead
        _key = key
        try:
            key(self._pb_body[0])
        except:
            _key = overriden_key

        if top_k is None or top_k >= len(self._pb_body):
            self._pb_body.sort(key=_key, reverse=reverse)
        else:
            # heap based sorting
            # makes sure that the top_k elements are correctly sorted and leaves the rest unsorted
            if _key is None:
                # heap is a list of documents
                heap = [element for element in self._pb_body]
            else:
                # heap is a list of tuples (key, document)
                heap = [
                    (_key(element), i, element)
                    for i, element in enumerate(self._pb_body)
                ]
            # if reverse use the maxheap operations for .heapify and .heappop
            heapify = heapq._heapify_max if reverse else heapq.heapify
            heappop = heapq._heappop_max if reverse else heapq.heappop

            # transform the original list to a heap and pop the top k elements
            heapify(heap)
            topk = [heappop(heap) for _ in range(top_k)]

            # get back to lists of docs from the lists of tuples
            _, _, topk = zip(*topk)
            _, _, heap = zip(*heap)
            topk, heap = list(topk), list(heap)
            # update the protobuf body
            self._pb_body = topk + heap

        self._update_id_to_index_map()

    @staticmethod
    def _flatten(sequence) -> 'DocumentArray':
        return DocumentArray(list(itertools.chain.from_iterable(sequence)))



import functools


class VersionedMixin:
    """
    Helper class to add versioning to an object. The version number is incremented each time an attribute is set.
    """

    _version = 0
    _ON_GETATTR = ['matches', 'chunks']

    def _increase_version(self):
        super().__setattr__('_version', self._version + 1)

    def __setattr__(self, attr, value):
        super().__setattr__(attr, value)
        self._increase_version()

    def __delattr__(self, attr):
        super(VersionedMixin, self).__delattr__(attr)
        self._increase_version()


def versioned(fn):
    """
    Decorator function that increases the version number each time the decorated method is called.
    The class of the decorated method must be a subclass of :class:`VersionedMixin`
    :param fn: the method to decorate
    :return: decorated function
    """

    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        self._increase_version()
        return fn(self, *args, **kwargs)

    return wrapper



import numpy as np

from ....helper import T, deprecate_by


class MeshDataMixin:
    """Provide helper functions for :class:`Document` to support 3D mesh data and point cloud. """

    def load_uri_to_point_cloud_blob(
        self: T, samples: int, as_chunks: bool = False
    ) -> T:
        """Convert a 3d mesh-like :attr:`.uri` into :attr:`.blob`

        :param samples: number of points to sample from the mesh
        :param as_chunks: when multiple geometry stored in one mesh file,
            then store each geometry into different :attr:`.chunks`

        :return: itself after processed
        """
        import trimesh

        mesh = trimesh.load_mesh(self.uri).deduplicated()

        pcs = []
        for geo in mesh.geometry.values():
            geo: trimesh.Trimesh
            pcs.append(geo.sample(samples))

        if as_chunks:
            from . import Document

            for p in pcs:
                self.chunks.append(Document(blob=p))
        else:
            self.blob = np.stack(pcs).squeeze()
        return self

    convert_uri_to_point_cloud_blob = deprecate_by(load_uri_to_point_cloud_blob)



import re
from typing import Dict, List, Optional, TYPE_CHECKING

from .... import __default_endpoint__
from ....excepts import (
    ExecutorFailToLoad,
    BadConfigSource,
)
from ....executors import BaseExecutor
from ....types.arrays.document import DocumentArray
from ....types.arrays.memmap import DocumentArrayMemmap
from ....types.message import Message, Request

if TYPE_CHECKING:
    import argparse
    from ....logging.logger import JinaLogger


def _get_docs_matrix_from_message(
    msg: 'Message',
    partial_request: Optional[List[Request]],
    field: str,
) -> List['DocumentArray']:
    if partial_request is not None:
        result = [getattr(r, field) for r in reversed(partial_request)]
    else:
        result = [getattr(msg.request, field)]

    # to unify all length=0 DocumentArray (or any other results) will simply considered as None
    # otherwise, the executor has to handle [None, None, None] or [DocArray(0), DocArray(0), DocArray(0)]
    len_r = sum(len(r) for r in result)
    if len_r:
        return result


def _get_docs_from_msg(
    msg: 'Message',
    partial_request: Optional[List[Request]],
    field: str,
) -> 'DocumentArray':
    if partial_request is not None:
        result = DocumentArray(
            [d for r in reversed(partial_request) for d in getattr(r, field)]
        )
    else:
        result = getattr(msg.request, field)

    return result


class DataRequestHandler:
    """Object to encapsulate the code related to handle the data requests passing to executor and its returned values"""

    def __init__(self, args: 'argparse.Namespace', logger: 'JinaLogger', **kwargs):
        """Initialize private parameters and execute private loading functions.

        :param args: args from CLI
        :param logger: the logger provided by the user
        :param kwargs: extra keyword arguments
        """
        super().__init__()
        self.args = args
        self.args.pea_id = self.args.shard_id
        self.args.parallel = self.args.shards
        self.logger = logger
        self._is_closed = False
        self._load_executor()

    def _load_executor(self):
        """Load the executor to this runtime, specified by ``uses`` CLI argument."""
        try:
            self._executor = BaseExecutor.load_config(
                self.args.uses,
                uses_with=self.args.uses_with,
                uses_metas=self.args.uses_metas,
                uses_requests=self.args.uses_requests,
                runtime_args=vars(self.args),
                extra_search_paths=self.args.extra_search_paths,
            )
        except BadConfigSource as ex:
            self.logger.error(
                f'fail to load config from {self.args.uses}, if you are using docker image for --uses, '
                f'please use "docker://YOUR_IMAGE_NAME"'
            )
            raise ExecutorFailToLoad from ex
        except FileNotFoundError as ex:
            self.logger.error(f'fail to load file dependency')
            raise ExecutorFailToLoad from ex
        except Exception as ex:
            self.logger.critical(f'can not load the executor from {self.args.uses}')
            raise ExecutorFailToLoad from ex

    @staticmethod
    def _parse_params(parameters: Dict, executor_name: str):
        parsed_params = parameters
        specific_parameters = parameters.get(executor_name, None)
        if specific_parameters:
            parsed_params.update(**specific_parameters)
        return parsed_params

    def handle(
        self,
        msg: 'Message',
        partial_requests: Optional[List[Request]],
        peapod_name: str,
    ):
        """Initialize private parameters and execute private loading functions.

        :param msg: The message to handle containing a DataRequest
        :param partial_requests: All the partial requests, to be considered when more than one expected part
        :param peapod_name: the name of the peapod owning this handler
        """
        # skip executor if target_peapod mismatch
        if not re.match(msg.envelope.header.target_peapod, peapod_name):
            self.logger.debug(
                f'skip executor: mismatch target, target: {msg.envelope.header.target_peapod}, name: {peapod_name}'
            )
            return

        # skip executor if endpoints mismatch
        if (
            msg.envelope.header.exec_endpoint not in self._executor.requests
            and __default_endpoint__ not in self._executor.requests
        ):
            self.logger.debug(
                f'skip executor: mismatch request, exec_endpoint: {msg.envelope.header.exec_endpoint}, requests: {self._executor.requests}'
            )
            if partial_requests:
                DataRequestHandler.replace_docs(
                    msg,
                    docs=_get_docs_from_msg(
                        msg,
                        partial_request=partial_requests,
                        field='docs',
                    ),
                )
            return

        params = self._parse_params(
            msg.request.parameters.to_dict(), self._executor.metas.name
        )
        docs = _get_docs_from_msg(
            msg,
            partial_request=partial_requests,
            field='docs',
        )
        # executor logic
        r_docs = self._executor(
            req_endpoint=msg.envelope.header.exec_endpoint,
            docs=docs,
            parameters=params,
            docs_matrix=_get_docs_matrix_from_message(
                msg,
                partial_request=partial_requests,
                field='docs',
            ),
            groundtruths=_get_docs_from_msg(
                msg,
                partial_request=partial_requests,
                field='groundtruths',
            ),
            groundtruths_matrix=_get_docs_matrix_from_message(
                msg,
                partial_request=partial_requests,
                field='groundtruths',
            ),
        )

        # assigning result back to request
        # 1. Return none: do nothing
        # 2. Return nonempty and non-DocumentArray: raise error
        # 3. Return DocArray, but the memory pointer says it is the same as self.docs: do nothing
        # 4. Return DocArray and its not a shallow copy of self.docs: assign self.request.docs
        if r_docs is not None:
            if isinstance(r_docs, (DocumentArray, DocumentArrayMemmap)):
                if r_docs != msg.request.docs:
                    # this means the returned DocArray is a completely new one
                    DataRequestHandler.replace_docs(msg, r_docs)
            elif isinstance(r_docs, dict):
                msg.request.parameters.update(r_docs)
            else:
                raise TypeError(
                    f'The return type must be DocumentArray / DocumentArrayMemmap / Dict / `None`, '
                    f'but getting {r_docs!r}'
                )
        elif partial_requests:
            DataRequestHandler.replace_docs(msg, docs)

    @staticmethod
    def replace_docs(msg, docs):
        """Replaces the docs in a message with new Documents.

        :param msg: The message object
        :param docs: the new docs to be used
        """
        msg.request.docs.clear()
        msg.request.docs.extend(docs)

    def close(self):
        """ Close the data request handler, by closing the executor """
        if not self._is_closed:
            self._executor.close()
            self._is_closed = True



"""This modules defines all kinds of exceptions raised in Jina."""


class BaseJinaExeception:
    """A base class for all exceptions raised by Jina"""


class NoExplicitMessage(Exception, BaseJinaExeception):
    """Waiting until all partial messages are received."""


class ChainedPodException(Exception, BaseJinaExeception):
    """Chained exception from the last Pod."""


class MismatchedVersion(SystemError, BaseJinaExeception):
    """When the jina version info of the incoming message does not match the local Jina version."""


class ExecutorFailToLoad(SystemError, BaseJinaExeception):
    """When the executor can not be loaded in pea/pod."""


class RuntimeFailToStart(SystemError, BaseJinaExeception):
    """When pea/pod is failed to started."""


class ScalingFails(SystemError, BaseJinaExeception):
    """When scaling is unsuccessful for an Executor."""


class MemoryOverHighWatermark(Exception, BaseJinaExeception):
    """When the memory usage is over the defined high water mark."""


class NoAvailablePortError(Exception, BaseJinaExeception):
    """When no available random port could be found"""


class RuntimeTerminated(KeyboardInterrupt, BaseJinaExeception):
    """The event loop of BasePea ends."""


class UnknownControlCommand(RuntimeError, BaseJinaExeception):
    """The control command received can not be recognized."""


class FlowTopologyError(Exception, BaseJinaExeception):
    """Flow exception when the topology is ambiguous."""


class FlowMissingPodError(Exception, BaseJinaExeception):
    """Flow exception when a pod can not be found in the flow."""


class FlowBuildLevelError(Exception, BaseJinaExeception):
    """Flow exception when required build level is higher than the current build level."""


class BadConfigSource(FileNotFoundError, BaseJinaExeception):
    """The yaml config file is bad, not loadable or not exist."""


class BadClient(Exception, BaseJinaExeception):
    """A wrongly defined client, can not communicate with jina server correctly."""


class BadClientCallback(BadClient, BaseJinaExeception):
    """Error in the callback function on the client side."""


class BadClientInput(BadClient, BaseJinaExeception):
    """Error in the request generator function on the client side."""


class BadDocType(TypeError, BaseJinaExeception):
    """Exception when can not construct a document from the given data."""


class BadRequestType(TypeError, BaseJinaExeception):
    """Exception when can not construct a request object from given data."""


class BadNamedScoreType(TypeError, BaseJinaExeception):
    """Exception when can not construct a named score from the given data."""


class BadImageNameError(Exception, BaseJinaExeception):
    """Exception when an image name can not be found either local & remote"""


class BadYAMLVersion(Exception, BaseJinaExeception):
    """Exception when YAML config specifies a wrong version number."""


class DaemonConnectivityError(Exception, BaseJinaExeception):
    """Exception to raise when jina daemon is not reachable."""


class DaemonWorkspaceCreationFailed(Exception, BaseJinaExeception):
    """Exception to raise when jina daemon is not connectable."""


class DaemonPeaCreationFailed(Exception, BaseJinaExeception):
    """Exception to raise when jina daemon is not connectable."""


class NotSupportedError(Exception, BaseJinaExeception):
    """Exception when user accidentally using a retired argument."""


class RoutingTableCyclicError(Exception, BaseJinaExeception):
    """Raised when the routing graph has cycles."""


class RuntimeRunForeverEarlyError(Exception, BaseJinaExeception):
    """Raised when an error occurs when starting the run_forever of Runtime"""


class DockerVersionError(SystemError, BaseJinaExeception):
    """Raised when the docker version is incompatible"""


class DaemonInvalidDockerfile(FileNotFoundError, BaseJinaExeception):
    """Raised when invalid dockerfile is passed to JinaD"""



import io
import os.path
from contextlib import nullcontext
from typing import Union, BinaryIO, TYPE_CHECKING, Type

from ..... import __windows__
from .....helper import random_uuid

if TYPE_CHECKING:
    from .....helper import T


class BinaryIOMixin:
    """Save/load an array to a binary file. """

    @classmethod
    def load_binary(cls: Type['T'], file: Union[str, BinaryIO, bytes]) -> 'T':
        """Load array elements from a LZ4-compressed binary file.

        :param file: File or filename or serialized bytes where the data is stored.

        :return: a DocumentArray object
        """

        if hasattr(file, 'read'):
            file_ctx = nullcontext(file)
        elif isinstance(file, bytes):
            file_ctx = nullcontext(file)
        elif os.path.exists(file):
            file_ctx = open(file, 'rb')
        else:
            raise ValueError(f'unsupported input {file!r}')

        from ...document import Document
        import lz4.frame

        with file_ctx as fp:
            d = fp.read() if hasattr(fp, 'read') else fp
            d = lz4.frame.decompress(d)
            _len = len(random_uuid().bytes)
            _binary_delimiter = d[:_len]  # first get delimiter
            da = cls()
            da.extend(Document(od) for od in d[_len:].split(_binary_delimiter))
            return da

    def save_binary(self, file: Union[str, BinaryIO]) -> None:
        """Save array elements into a LZ4 compressed binary file.

        Comparing to :meth:`save_json`, it is faster and the file is smaller, but not human-readable.

        .. note::
            To get a binary presentation in memory, use ``bytes(...)``.

        :param file: File or filename to which the data is saved.
        """
        if hasattr(file, 'write'):
            file_ctx = nullcontext(file)
        else:
            if __windows__:
                file_ctx = open(file, 'wb', newline='')
            else:
                file_ctx = open(file, 'wb')

        with file_ctx as fp:
            fp.write(bytes(self))

    def to_bytes(self) -> bytes:
        """Serialize itself into bytes with LZ4 compression.

        For more Pythonic code, please use ``bytes(...)``.

        :return: the binary serialization in bytes
        """
        import lz4.frame

        _binary_delimiter = random_uuid().bytes
        with io.BytesIO() as bf:
            with lz4.frame.LZ4FrameFile(bf, 'wb') as f:
                for d in self:
                    f.write(_binary_delimiter)
                    f.write(bytes(d))
            return bf.getvalue()

    def __bytes__(self):
        return self.to_bytes()



ac_table = {
    'commands': [
        '--help',
        '--version',
        '--version-full',
        'hello',
        'executor',
        'flow',
        'ping',
        'gateway',
        'hub',
        'help',
        'pea',
        'pod',
        'client',
        'export-api',
    ],
    'completions': {
        'hello fashion': [
            '--help',
            '--workdir',
            '--download-proxy',
            '--index-data-url',
            '--index-labels-url',
            '--query-data-url',
            '--query-labels-url',
            '--num-query',
            '--top-k',
        ],
        'hello chatbot': [
            '--help',
            '--workdir',
            '--download-proxy',
            '--index-data-url',
            '--port-expose',
            '--replicas',
            '--unblock-query-flow',
        ],
        'hello multimodal': [
            '--help',
            '--workdir',
            '--download-proxy',
            '--index-data-url',
            '--port-expose',
            '--unblock-query-flow',
        ],
        'hello fork': ['--help', 'fashion', 'chatbot', 'multimodal'],
        'hello': ['--help', 'fashion', 'chatbot', 'multimodal', 'fork'],
        'executor': [
            '--help',
            '--name',
            '--workspace',
            '--k8s-namespace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--workspace-id',
            '--static-routing-table',
            '--routing-table',
            '--dynamic-routing',
            '--extra-search-paths',
            '--zmq-identity',
            '--port-ctrl',
            '--ctrl-with-ipc',
            '--timeout-ctrl',
            '--ssh-server',
            '--ssh-keyfile',
            '--ssh-password',
            '--uses',
            '--uses-with',
            '--uses-metas',
            '--uses-requests',
            '--py-modules',
            '--port-in',
            '--port-out',
            '--hosts-in-connect',
            '--host-in',
            '--host-out',
            '--socket-in',
            '--socket-out',
            '--memory-hwm',
            '--on-error-strategy',
            '--native',
            '--num-part',
            '--dynamic-routing-out',
            '--dynamic-routing-in',
            '--grpc-data-requests',
            '--runs-in-docker',
            '--dump-path',
            '--k8s-disable-connection-pool',
            '--entrypoint',
            '--docker-kwargs',
            '--pull-latest',
            '--volumes',
            '--gpus',
            '--host',
            '--port-jinad',
            '--quiet-remote-logs',
            '--upload-files',
            '--disable-remote',
            '--daemon',
            '--runtime-backend',
            '--runtime',
            '--runtime-cls',
            '--timeout-ready',
            '--env',
            '--expose-public',
            '--shard-id',
            '--pea-id',
            '--replica-id',
            '--pea-role',
            '--noblock-on-start',
            '--shards',
            '--parallel',
            '--replicas',
            '--install-requirements',
            '--force-update',
            '--force',
        ],
        'flow': [
            '--help',
            '--name',
            '--workspace',
            '--k8s-namespace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--workspace-id',
            '--static-routing-table',
            '--routing-table',
            '--dynamic-routing',
            '--extra-search-paths',
            '--uses',
            '--env',
            '--inspect',
            '--infrastructure',
        ],
        'ping': ['--help', '--timeout', '--retries', '--print-response'],
        'gateway': [
            '--help',
            '--name',
            '--workspace',
            '--k8s-namespace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--workspace-id',
            '--static-routing-table',
            '--routing-table',
            '--dynamic-routing',
            '--extra-search-paths',
            '--zmq-identity',
            '--port-ctrl',
            '--ctrl-with-ipc',
            '--timeout-ctrl',
            '--ssh-server',
            '--ssh-keyfile',
            '--ssh-password',
            '--uses',
            '--uses-with',
            '--uses-metas',
            '--uses-requests',
            '--py-modules',
            '--port-in',
            '--port-out',
            '--hosts-in-connect',
            '--host-in',
            '--host-out',
            '--socket-in',
            '--socket-out',
            '--memory-hwm',
            '--on-error-strategy',
            '--native',
            '--num-part',
            '--dynamic-routing-out',
            '--dynamic-routing-in',
            '--grpc-data-requests',
            '--runs-in-docker',
            '--dump-path',
            '--k8s-disable-connection-pool',
            '--prefetch',
            '--title',
            '--description',
            '--cors',
            '--default-swagger-ui',
            '--no-debug-endpoints',
            '--no-crud-endpoints',
            '--expose-endpoints',
            '--uvicorn-kwargs',
            '--compress',
            '--compress-min-bytes',
            '--compress-min-ratio',
            '--protocol',
            '--host',
            '--proxy',
            '--port-expose',
            '--daemon',
            '--runtime-backend',
            '--runtime',
            '--runtime-cls',
            '--timeout-ready',
            '--env',
            '--expose-public',
            '--shard-id',
            '--pea-id',
            '--replica-id',
            '--pea-role',
            '--noblock-on-start',
            '--shards',
            '--parallel',
            '--replicas',
            '--k8s-uses-init',
            '--k8s-mount-path',
            '--k8s-init-container-command',
            '--k8s-custom-resource-dir',
        ],
        'hub new': [
            '--help',
            '--name',
            '--path',
            '--advance-configuration',
            '--description',
            '--keywords',
            '--url',
            '--add-dockerfile',
        ],
        'hub push': [
            '--help',
            '--no-usage',
            '--dockerfile',
            '--tag',
            '--force-update',
            '--force',
            '--secret',
            '--public',
            '--private',
        ],
        'hub pull': [
            '--help',
            '--no-usage',
            '--install-requirements',
            '--force-update',
            '--force',
        ],
        'hub': ['--help', 'new', 'push', 'pull'],
        'help': ['--help'],
        'pea': [
            '--help',
            '--name',
            '--workspace',
            '--k8s-namespace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--workspace-id',
            '--static-routing-table',
            '--routing-table',
            '--dynamic-routing',
            '--extra-search-paths',
            '--zmq-identity',
            '--port-ctrl',
            '--ctrl-with-ipc',
            '--timeout-ctrl',
            '--ssh-server',
            '--ssh-keyfile',
            '--ssh-password',
            '--uses',
            '--uses-with',
            '--uses-metas',
            '--uses-requests',
            '--py-modules',
            '--port-in',
            '--port-out',
            '--hosts-in-connect',
            '--host-in',
            '--host-out',
            '--socket-in',
            '--socket-out',
            '--memory-hwm',
            '--on-error-strategy',
            '--native',
            '--num-part',
            '--dynamic-routing-out',
            '--dynamic-routing-in',
            '--grpc-data-requests',
            '--runs-in-docker',
            '--dump-path',
            '--k8s-disable-connection-pool',
            '--entrypoint',
            '--docker-kwargs',
            '--pull-latest',
            '--volumes',
            '--gpus',
            '--host',
            '--port-jinad',
            '--quiet-remote-logs',
            '--upload-files',
            '--disable-remote',
            '--daemon',
            '--runtime-backend',
            '--runtime',
            '--runtime-cls',
            '--timeout-ready',
            '--env',
            '--expose-public',
            '--shard-id',
            '--pea-id',
            '--replica-id',
            '--pea-role',
            '--noblock-on-start',
            '--shards',
            '--parallel',
            '--replicas',
            '--install-requirements',
            '--force-update',
            '--force',
        ],
        'pod': [
            '--help',
            '--name',
            '--workspace',
            '--k8s-namespace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--workspace-id',
            '--static-routing-table',
            '--routing-table',
            '--dynamic-routing',
            '--extra-search-paths',
            '--zmq-identity',
            '--port-ctrl',
            '--ctrl-with-ipc',
            '--timeout-ctrl',
            '--ssh-server',
            '--ssh-keyfile',
            '--ssh-password',
            '--uses',
            '--uses-with',
            '--uses-metas',
            '--uses-requests',
            '--py-modules',
            '--port-in',
            '--port-out',
            '--hosts-in-connect',
            '--host-in',
            '--host-out',
            '--socket-in',
            '--socket-out',
            '--memory-hwm',
            '--on-error-strategy',
            '--native',
            '--num-part',
            '--dynamic-routing-out',
            '--dynamic-routing-in',
            '--grpc-data-requests',
            '--runs-in-docker',
            '--dump-path',
            '--k8s-disable-connection-pool',
            '--entrypoint',
            '--docker-kwargs',
            '--pull-latest',
            '--volumes',
            '--gpus',
            '--host',
            '--port-jinad',
            '--quiet-remote-logs',
            '--upload-files',
            '--disable-remote',
            '--daemon',
            '--runtime-backend',
            '--runtime',
            '--runtime-cls',
            '--timeout-ready',
            '--env',
            '--expose-public',
            '--shard-id',
            '--pea-id',
            '--replica-id',
            '--pea-role',
            '--noblock-on-start',
            '--shards',
            '--parallel',
            '--replicas',
            '--install-requirements',
            '--force-update',
            '--force',
            '--uses-before',
            '--uses-after',
            '--polling',
            '--scheduling',
            '--external',
            '--peas-hosts',
            '--pod-role',
            '--no-dynamic-routing',
            '--connect-to-predecessor',
            '--k8s-uses-init',
            '--k8s-mount-path',
            '--k8s-init-container-command',
            '--k8s-custom-resource-dir',
        ],
        'client': [
            '--help',
            '--host',
            '--proxy',
            '--port',
            '--https',
            '--asyncio',
            '--protocol',
        ],
        'export-api': ['--help', '--yaml-path', '--json-path', '--schema-path'],
    },
}



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import serializer as jina__pb2


class JinaRPCStub(object):
    """*
    jina gRPC service.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Call = channel.stream_stream(
            '/jina.JinaRPC/Call',
            request_serializer=jina__pb2.RequestProto.SerializeToString,
            response_deserializer=jina__pb2.RequestProto.FromString,
        )


class JinaRPCServicer(object):
    """*
    jina gRPC service.
    """

    def Call(self, request_iterator, context):
        """Pass in a Request and a filled Request with matches will be returned."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_JinaRPCServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'Call': grpc.stream_stream_rpc_method_handler(
            servicer.Call,
            request_deserializer=jina__pb2.RequestProto.FromString,
            response_serializer=jina__pb2.RequestProto.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'jina.JinaRPC', rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class JinaRPC(object):
    """*
    jina gRPC service.
    """

    @staticmethod
    def Call(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/jina.JinaRPC/Call',
            jina__pb2.RequestProto.SerializeToString,
            jina__pb2.RequestProto.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )


class JinaDataRequestRPCStub(object):
    """*
    jina gRPC service for DataRequests.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Call = channel.unary_unary(
            '/jina.JinaDataRequestRPC/Call',
            request_serializer=jina__pb2.MessageProto.SerializeToString,
            response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )


class JinaDataRequestRPCServicer(object):
    """*
    jina gRPC service for DataRequests.
    """

    def Call(self, request, context):
        """Pass in a Message, wrapping a DataRequest"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_JinaDataRequestRPCServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'Call': grpc.unary_unary_rpc_method_handler(
            servicer.Call,
            request_deserializer=jina__pb2.MessageProto.FromString,
            response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'jina.JinaDataRequestRPC', rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class JinaDataRequestRPC(object):
    """*
    jina gRPC service for DataRequests.
    """

    @staticmethod
    def Call(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jina.JinaDataRequestRPC/Call',
            jina__pb2.MessageProto.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# some implementations are adopted from https://github.com/ncoop57/cute_ranking/blob/main/cute_ranking/core.py
# the original code is licensed under Apache-2.0

from typing import List, Optional

import numpy as np


def _check_k(k):
    if k is not None and k < 1:
        raise ValueError(f'`k` must be >=1 or `None`')


def r_precision(binary_relevance: List[int]) -> float:
    """R Precision after all relevant documents have been retrieved
    Relevance is binary (nonzero is relevant).

    .. seealso::
        https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#R-precision

    :param binary_relevance: binary relevancy in rank order
    :return: precision
    """
    binary_relevance = np.array(binary_relevance) != 0
    z = binary_relevance.nonzero()[0]
    if not z.size:
        return 0.0
    return float(np.mean(binary_relevance[: z[-1] + 1]))


def precision_at_k(binary_relevance: List[int], k: Optional[int] = None) -> float:
    """Precision @K.

    :param binary_relevance: binary relevancy in rank order
    :param k: measured on top-k
    :return: precision @k
    """
    _check_k(k)
    binary_relevance = np.array(binary_relevance)[:k] != 0
    return float(np.mean(binary_relevance))


def hit_at_k(binary_relevance: List[int], k: Optional[int] = None) -> int:
    """Score is percentage of first relevant item in list that occur

    :param binary_relevance: binary relevancy in rank order
    :param k: measured on top-k
    :return: hit @k if hit return 1 else 0
    """
    _check_k(k)
    return 1 if np.sum(binary_relevance[:k]) > 0 else 0


def average_precision(binary_relevance: List[int]) -> float:
    """Score is average precision (area under PR curve)
    Relevance is binary (nonzero is relevant).

    :param binary_relevance: binary relevancy in rank order
    :return: Average precision
    """
    r = np.array(binary_relevance) != 0
    out = [precision_at_k(r, k + 1) for k in range(r.size) if r[k]]
    if not out:
        return 0.0
    return float(np.mean(out))


def reciprocal_rank(binary_relevance: List[int]) -> float:
    """Score is reciprocal of the rank of the first relevant item

    :param binary_relevance: binary relevancy in rank order
    :return: Average precision
    """
    rs = np.array(binary_relevance).nonzero()[0]
    return 1.0 / (rs[0] + 1) if rs.size else 0.0


def recall_at_k(
    binary_relevance: List[int], max_rel: int, k: Optional[int] = None
) -> float:
    """Score is recall after all relevant documents have been retrieved
    Relevance is binary (nonzero is relevant).

    :param binary_relevance: binary relevancy in rank order
    :param k: measured on top-k
    :param max_rel: Maximum number of documents that can be relevant
    :return: Recall score
    """
    _check_k(k)
    binary_relevance = np.array(binary_relevance[:k]) != 0
    if np.sum(binary_relevance) > max_rel:
        raise ValueError(f'Number of relevant Documents retrieved > {max_rel}')
    return np.sum(binary_relevance) / max_rel


def f1_score_at_k(
    binary_relevance: List[int], max_rel: int, k: Optional[int] = None
) -> float:
    """Score is harmonic mean of precision and recall
    Relevance is binary (nonzero is relevant).

    :param binary_relevance: binary relevancy in rank order
    :param k: measured on top-k
    :param max_rel: Maximum number of documents that can be relevant
    :return: F1 score @ k
    """
    _check_k(k)
    p = precision_at_k(binary_relevance, k)
    r = recall_at_k(binary_relevance, max_rel, k)
    if (p + r) > 0:
        return 2 * p * r / (p + r)
    else:
        return 0.0


def dcg_at_k(relevance: List[float], method: int = 0, k: Optional[int] = None):
    """Score is discounted cumulative gain (dcg)
    Relevance is positive real values. Can use binary
    as the previous methods.

    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf

    :param relevance: Relevance scores (list or numpy) in rank order
            (first element is the first item)
    :param k: measured on top-k
    :param method: If 0 then weights are [1.0, 1.0, 0.6309, 0.5, 0.4307, ...]
                If 1 then weights are [1.0, 0.6309, 0.5, 0.4307, ...]
    :return: Discounted cumulative gain
    """
    _check_k(k)
    r = np.asfarray(relevance)[:k]
    if r.size:
        if method == 0:
            return r[0] + np.sum(r[1:] / np.log2(np.arange(2, r.size + 1)))
        elif method == 1:
            return np.sum(r / np.log2(np.arange(2, r.size + 2)))
        else:
            raise ValueError('method must be 0 or 1.')
    return 0.0


def ndcg_at_k(relevance: List[float], method: int = 0, k: Optional[int] = None):
    """Score is normalized discounted cumulative gain (ndcg)
    Relevance is positive real values.  Can use binary
    as the previous methods.

    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf

    :param relevance: Relevance scores (list or numpy) in rank order
            (first element is the first item)
    :param k: measured on top-k
    :param method: If 0 then weights are [1.0, 1.0, 0.6309, 0.5, 0.4307, ...]
                If 1 then weights are [1.0, 0.6309, 0.5, 0.4307, ...]
    :return: Normalized discounted cumulative gain
    """
    _check_k(k)
    dcg_max = dcg_at_k(sorted(relevance, reverse=True), method=method, k=k)
    if not dcg_max:
        return 0.0
    return dcg_at_k(relevance, method=method, k=k) / dcg_max



import asyncio
from contextlib import nullcontext
from typing import TYPE_CHECKING, Optional

import grpc

from ..base import BaseClient
from ..helper import callback_exec
from ...excepts import BadClient, BadClientInput
from ...logging.profile import ProgressBar
from ...proto import jina_pb2_grpc

if TYPE_CHECKING:
    from . import InputType, CallbackFnType


class GRPCBaseClient(BaseClient):
    """A simple Python client for connecting to the gRPC gateway.

    It manages the asyncio event loop internally, so all interfaces are synchronous from the outside.
    """

    async def _get_results(
        self,
        inputs: 'InputType',
        on_done: 'CallbackFnType',
        on_error: Optional['CallbackFnType'] = None,
        on_always: Optional['CallbackFnType'] = None,
        **kwargs,
    ):
        try:
            self.inputs = inputs
            req_iter = self._get_requests(**kwargs)
            async with grpc.aio.insecure_channel(
                f'{self.args.host}:{self.args.port}',
                options=[
                    ('grpc.max_send_message_length', -1),
                    ('grpc.max_receive_message_length', -1),
                ],
            ) as channel:
                stub = jina_pb2_grpc.JinaRPCStub(channel)
                self.logger.debug(f'connected to {self.args.host}:{self.args.port}')

                cm1 = (
                    ProgressBar(total_length=self._inputs_length)
                    if self.show_progress
                    else nullcontext()
                )

                with cm1 as p_bar:
                    async for resp in stub.Call(req_iter):
                        resp.as_typed_request(resp.request_type)
                        resp = resp.as_response()
                        callback_exec(
                            response=resp,
                            on_error=on_error,
                            on_done=on_done,
                            on_always=on_always,
                            continue_on_error=self.continue_on_error,
                            logger=self.logger,
                        )
                        if self.show_progress:
                            p_bar.update()
                        yield resp
        except KeyboardInterrupt:
            self.logger.warning('user cancel the process')
        except asyncio.CancelledError as ex:
            self.logger.warning(f'process error: {ex!r}')
        except grpc.aio._call.AioRpcError as rpc_ex:
            # Since this object is guaranteed to be a grpc.Call, might as well include that in its name.
            my_code = rpc_ex.code()
            my_details = rpc_ex.details()
            msg = f'gRPC error: {my_code} {my_details}'
            if my_code == grpc.StatusCode.UNAVAILABLE:
                self.logger.error(
                    f'{msg}\nthe ongoing request is terminated as the server is not available or closed already'
                )
                raise rpc_ex
            elif my_code == grpc.StatusCode.INTERNAL:
                self.logger.error(f'{msg}\ninternal error on the server side')
                raise rpc_ex
            elif (
                my_code == grpc.StatusCode.UNKNOWN
                and 'asyncio.exceptions.TimeoutError' in my_details
            ):
                raise BadClientInput(
                    f'{msg}\n'
                    'often the case is that you define/send a bad input iterator to jina, '
                    'please double check your input iterator'
                ) from rpc_ex
            else:
                raise BadClient(msg) from rpc_ex



import json
import os.path
import shutil
import tempfile
import threading
import warnings
from math import sqrt, ceil, floor
from typing import Optional

import numpy as np

from .... import Document, __resources_path__
from ....helper import random_port


class PlotMixin:
    """Helper functions for plotting the arrays. """

    def plot_embeddings(
        self,
        title: str = 'MyDocumentArray',
        path: Optional[str] = None,
        image_sprites: bool = False,
        min_image_size: int = 16,
        channel_axis: int = -1,
        start_server: bool = True,
        port: Optional[int] = None,
    ) -> str:
        """Interactively visualize :attr:`.embeddings` using the Embedding Projector.

        :param title: the title of this visualization. If you want to compare multiple embeddings at the same time,
                make sure to give different names each time and set ``path`` to the same value.
        :param port: if set, run the embedding-projector frontend at given port. Otherwise a random port is used.
        :param image_sprites: if set, visualize the dots using :attr:`.uri` and :attr:`.blob`.
        :param path: if set, then append the visualization to an existing folder, where you can compare multiple
            embeddings at the same time. Make sure to use a different ``title`` each time .
        :param min_image_size: only used when `image_sprites=True`. the minimum size of the image
        :param channel_axis: only used when `image_sprites=True`. the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param start_server: if set, start a HTTP server and open the frontend directly. Otherwise, you need to rely on ``return`` path and serve by yourself.
        :return: the path to the embeddings visualization info.
        """
        path = path or tempfile.mkdtemp()
        emb_fn = f'{title}.tsv'
        meta_fn = f'{title}.metas.tsv'
        config_fn = f'config.json'
        sprite_fn = f'{title}.png'

        if image_sprites:
            img_per_row = ceil(sqrt(len(self)))
            canvas_size = min(img_per_row * min_image_size, 8192)
            img_size = max(int(canvas_size / img_per_row), min_image_size)

            max_docs = ceil(canvas_size / img_size) ** 2
            if len(self) > max_docs:
                warnings.warn(
                    f'''
                    {self!r} has more than {max_docs} elements, which is the maximum number of image sprites can support. 
                    The resulting visualization may not be correct. You can do the following:
                    
                    - use fewer images: `da[:10000].plot_embeddings()`
                    - reduce the `min_image_size` to a smaller number, say 8 or 4 (but bear in mind you can hardly recognize anything with a 4x4 image)
                    - turn off `image_sprites` via `da.plot_embeddings(image_sprites=False)`
                    '''
                )

            self.plot_image_sprites(
                os.path.join(path, sprite_fn),
                canvas_size=canvas_size,
                min_size=min_image_size,
                channel_axis=channel_axis,
            )

        self.save_embeddings_csv(os.path.join(path, emb_fn), delimiter='\t')
        self.save_csv(
            os.path.join(path, meta_fn),
            exclude_fields=('embedding', 'blob', 'scores'),
            dialect='excel-tab',
        )

        _epj_config = {
            'embeddings': [
                {
                    'tensorName': title,
                    'tensorShape': list(self.embeddings.shape),
                    'tensorPath': f'/static/{emb_fn}',
                    'metadataPath': f'/static/{meta_fn}',
                    'sprite': {
                        'imagePath': f'/static/{sprite_fn}',
                        'singleImageDim': (img_size,) * 2,
                    }
                    if image_sprites
                    else {},
                }
            ]
        }

        if os.path.exists(os.path.join(path, config_fn)):
            with open(os.path.join(path, config_fn)) as fp:
                old_config = json.load(fp)
                _epj_config['embeddings'].extend(old_config.get('embeddings', []))

        with open(os.path.join(path, config_fn), 'w') as fp:
            json.dump(_epj_config, fp)

        shutil.copyfile(
            os.path.join(__resources_path__, 'embedding-projector/index.html'),
            os.path.join(path, 'index.html'),
        )

        if start_server:

            def _get_fastapi_app():
                from fastapi import FastAPI
                from starlette.middleware.cors import CORSMiddleware
                from starlette.staticfiles import StaticFiles

                app = FastAPI()
                app.add_middleware(
                    CORSMiddleware,
                    allow_origins=['*'],
                    allow_credentials=True,
                    allow_methods=['*'],
                    allow_headers=['*'],
                )
                app.mount('/static', StaticFiles(directory=path), name='static')
                return app

            import uvicorn

            app = _get_fastapi_app()
            port = port or random_port()
            t_m = threading.Thread(
                target=uvicorn.run,
                kwargs=dict(app=app, port=port, log_level='error'),
                daemon=True,
            )
            t_m.start()
            url_html_path = (
                f'http://localhost:{port}/static/index.html?config={config_fn}'
            )
            try:
                import webbrowser

                webbrowser.open(url_html_path, new=2)
            except:
                pass  # intentional pass, browser support isn't cross-platform
            finally:
                from ....logging.predefined import default_logger

                default_logger.info(
                    f'You should see a webpage opened in your browser, '
                    f'if not, you may open {url_html_path} manually'
                )
            t_m.join()
        return path

    def plot_embeddings_legacy(
        self,
        output: Optional[str] = None,
        title: Optional[str] = None,
        colored_attr: Optional[str] = None,
        colormap: str = 'rainbow',
        method: str = 'pca',
        show_axis: bool = False,
        **kwargs,
    ):
        """Plot embeddings in a 2D projection with the PCA algorithm. This function requires ``matplotlib`` installed.

        If `tag_name` is provided the plot uses a distinct color for each unique tag value in the
        documents of the DocumentArray.

        :param output: Optional path to store the visualization. If not given, show in UI
        :param title: Optional title of the plot. When not given, the default title is used.
        :param colored_attr: Optional str that specifies attribute used to color the plot, it supports dunder expression
            such as `tags__label`, `matches__0__id`.
        :param colormap: the colormap string supported by matplotlib.
        :param method: the visualization method, available `pca`, `tsne`. `pca` is fast but may not well represent
                nonlinear relationship of high-dimensional data. `tsne` requires scikit-learn to be installed and is
                much slower.
        :param show_axis: If set, axis and bounding box of the plot will be printed.
        :param kwargs: extra kwargs pass to matplotlib.plot
        """

        x_mat = self.embeddings
        assert isinstance(
            x_mat, np.ndarray
        ), f'Type {type(x_mat)} not currently supported, use np.ndarray embeddings'

        if method == 'tsne':
            from sklearn.manifold import TSNE

            x_mat_2d = TSNE(n_components=2).fit_transform(x_mat)
        else:
            from ....math.dimensionality_reduction import PCA

            x_mat_2d = PCA(n_components=2).fit_transform(x_mat)

        plt_kwargs = {
            'x': x_mat_2d[:, 0],
            'y': x_mat_2d[:, 1],
            'alpha': 0.2,
            'marker': '.',
        }

        import matplotlib.pyplot as plt

        plt.figure(figsize=(8, 8))

        plt.title(title or f'{len(x_mat)} Documents with {method}')

        if colored_attr:
            tags = []

            for x in self:
                try:
                    tags.append(getattr(x, colored_attr))
                except (KeyError, AttributeError, ValueError):
                    tags.append(None)
            tag_to_num = {tag: num for num, tag in enumerate(set(tags))}
            plt_kwargs['c'] = np.array([tag_to_num[ni] for ni in tags])
            plt_kwargs['cmap'] = plt.get_cmap(colormap)

        # update the plt_kwargs
        plt_kwargs.update(kwargs)

        plt.scatter(**plt_kwargs)

        if not show_axis:
            plt.gca().set_axis_off()
            plt.gca().xaxis.set_major_locator(plt.NullLocator())
            plt.gca().yaxis.set_major_locator(plt.NullLocator())

        if output:
            plt.savefig(output, bbox_inches='tight', pad_inches=0.1)
        else:
            plt.show()

    def plot_image_sprites(
        self,
        output: Optional[str] = None,
        canvas_size: int = 512,
        min_size: int = 16,
        channel_axis: int = -1,
    ) -> None:
        """Generate a sprite image for all image blobs in this DocumentArray-like object.

        An image sprite is a collection of images put into a single image. It is always square-sized.
        Each sub-image is also square-sized and equally-sized.

        :param output: Optional path to store the visualization. If not given, show in UI
        :param canvas_size: the size of the canvas
        :param min_size: the minimum size of the image
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        """
        if not self:
            raise ValueError(f'{self!r} is empty')

        import matplotlib.pyplot as plt

        img_per_row = ceil(sqrt(len(self)))
        img_size = int(canvas_size / img_per_row)

        if img_size < min_size:
            # image is too small, recompute the size
            img_size = min_size
            img_per_row = int(canvas_size / img_size)

        max_num_img = img_per_row ** 2
        sprite_img = np.zeros(
            [img_size * img_per_row, img_size * img_per_row, 3], dtype='uint8'
        )
        img_id = 0

        for d in self:
            _d = Document(d, copy=True)
            if _d.content_type != 'blob':
                _d.load_uri_to_image_blob()
                channel_axis = -1

            _d.set_image_blob_channel_axis(channel_axis, -1).set_image_blob_shape(
                shape=(img_size, img_size)
            )

            row_id = floor(img_id / img_per_row)
            col_id = img_id % img_per_row
            sprite_img[
                (row_id * img_size) : ((row_id + 1) * img_size),
                (col_id * img_size) : ((col_id + 1) * img_size),
            ] = _d.blob

            img_id += 1
            if img_id >= max_num_img:
                break

        from PIL import Image

        im = Image.fromarray(sprite_img)

        if output:
            with open(output, 'wb') as fp:
                im.save(fp)
        else:
            plt.gca().set_axis_off()
            plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
            plt.margins(0, 0)
            plt.gca().xaxis.set_major_locator(plt.NullLocator())
            plt.gca().yaxis.set_major_locator(plt.NullLocator())
            plt.imshow(im)
            plt.show()



from typing import Union, BinaryIO

import numpy as np

from ....helper import T, deprecate_by


class VideoDataMixin:
    """Provide helper functions for :class:`Document` to support video data. """

    def load_uri_to_video_blob(self: T, only_keyframes: bool = False) -> T:
        """Convert a :attr:`.uri` to a video ndarray :attr:`.blob`.

        :param only_keyframes: only keep the keyframes in the video
        :return: Document itself after processed
        """
        import av

        with av.open(self.uri) as container:
            if only_keyframes:
                stream = container.streams.video[0]
                stream.codec_context.skip_frame = 'NONKEY'

            frames = []
            for frame in container.decode(video=0):
                img = frame.to_image()
                frames.append(np.asarray(img))

        self.blob = np.moveaxis(np.stack(frames), 1, 2)
        return self

    def dump_video_blob_to_file(
        self: T, file: Union[str, BinaryIO], frame_rate: int = 30, codec: str = 'h264'
    ) -> T:
        """Save :attr:`.blob` as a video mp4/h264 file.

        :param file: The file to open, which can be either a string or a file-like object.
        :param frame_rate: frames per second
        :param codec: the name of a decoder/encoder
        :return: itself after processed
        """
        if (
            self.blob.ndim != 4
            or self.blob.shape[-1] != 3
            or self.blob.dtype != np.uint8
        ):
            raise ValueError(
                f'expects `.blob` with dtype=uint8 and ndim=4 and the last dimension is 3, '
                f'but receiving {self.blob.shape} in {self.blob.dtype}'
            )

        video_blob = np.moveaxis(np.clip(self.blob, 0, 255), 1, 2)

        import av

        with av.open(file, mode='w') as container:
            stream = container.add_stream(codec, rate=frame_rate)
            stream.width = self.blob.shape[1]
            stream.height = self.blob.shape[2]
            stream.pix_fmt = 'yuv420p'

            for b in video_blob:
                frame = av.VideoFrame.from_ndarray(b, format='rgb24')
                for packet in stream.encode(frame):
                    container.mux(packet)

            for packet in stream.encode():
                container.mux(packet)
        return self

    convert_uri_to_video_blob = deprecate_by(load_uri_to_video_blob)



from .helper import _uri_to_buffer, _to_datauri
from ....helper import T, deprecate_by


class BufferDataMixin:
    """Provide helper functions for :class:`Document` to handle binary data. """

    def load_uri_to_buffer(self: T) -> T:
        """Convert :attr:`.uri` to :attr:`.buffer` inplace.
        Internally it downloads from the URI and set :attr:`buffer`.

        :return: itself after processed
        """
        self.buffer = _uri_to_buffer(self.uri)
        return self

    def dump_buffer_to_datauri(
        self: T, charset: str = 'utf-8', base64: bool = False
    ) -> T:
        """Convert :attr:`.buffer` to data :attr:`.uri` in place.
        Internally it first reads into buffer and then converts it to data URI.

        :param charset: charset may be any character set registered with IANA
        :param base64: used to encode arbitrary octet sequences into a form that satisfies the rules of 7bit.
            Designed to be efficient for non-text 8 bit and binary data. Sometimes used for text data that
            frequently uses non-US-ASCII characters.

        :return: itself after processed
        """

        if not self.mime_type:
            raise ValueError(
                f'{self.mime_type} is unset, can not convert it to data uri'
            )

        self.uri = _to_datauri(
            self.mime_type, self.buffer, charset, base64, binary=True
        )
        return self

    convert_buffer_to_uri = deprecate_by(dump_buffer_to_datauri)
    convert_uri_to_buffer = deprecate_by(load_uri_to_buffer)



"""Module for helper functions for Hub API."""

import hashlib
import io
import json
import os
import shelve
import subprocess
import sys
import tarfile
import urllib
import warnings
import zipfile
from functools import lru_cache, wraps
from pathlib import Path
from typing import Tuple, Optional, Dict
from urllib.parse import urlparse, urljoin
from urllib.request import Request, urlopen
from contextlib import nullcontext

from .. import __resources_path__
from ..importer import ImportExtensions
from ..logging.predefined import default_logger


@lru_cache()
def get_hubble_url() -> str:
    """Get the Hubble URL from api.jina.ai or os.environ

    :return: the Hubble URL
    """
    if 'JINA_HUBBLE_REGISTRY' in os.environ:
        u = os.environ['JINA_HUBBLE_REGISTRY']
    else:
        try:
            req = Request(
                'https://api.jina.ai/hub/hubble.json',
                headers={'User-Agent': 'Mozilla/5.0'},
            )
            with urlopen(req) as resp:
                u = json.load(resp)['url']
        except:
            default_logger.critical(
                'Can not fetch the URL of Hubble from `api.jina.ai`'
            )
            raise
    return urljoin(u, '/v1/executors')


def parse_hub_uri(uri_path: str) -> Tuple[str, str, str, str]:
    """Parse the uri of the Jina Hub executor.

    :param uri_path: the uri of Jina Hub executor
    :return: a tuple of schema, id, tag, and secret
    """
    parser = urlparse(uri_path)
    scheme = parser.scheme
    if scheme not in {'jinahub', 'jinahub+docker'}:
        raise ValueError(f'{uri_path} is not a valid Hub URI.')

    items = list(parser.netloc.split(':'))
    name = items[0]

    if not name:
        raise ValueError(f'{uri_path} is not a valid Hub URI.')

    secret = items[1] if len(items) > 1 else None
    tag = parser.path.strip('/') if parser.path else None

    return scheme, name, tag, secret


def is_valid_huburi(uri: str) -> bool:
    """Return True if it is a valid Hubble URI

    :param uri: the uri to test
    :return: True or False
    """
    try:
        parse_hub_uri(uri)
        return True
    except:
        return False


def md5file(file_path: 'Path') -> str:
    """Retrun the MD5 checksum of the file

    :param file_path: the file to check md5sum
    :return: the MD5 checksum
    """
    hash_md5 = hashlib.md5()
    with file_path.open(mode='rb') as fp:
        for chunk in iter(lambda: fp.read(128 * hash_md5.block_size), b''):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()


def unpack_package(filepath: 'Path', target_dir: 'Path'):
    """Unpack the file to the target_dir.

    :param filepath: the path of given file
    :param target_dir: the path of target folder
    """
    if filepath.suffix == '.zip':
        with zipfile.ZipFile(filepath, 'r') as zip:
            zip.extractall(target_dir)
    elif filepath.suffix in ['.tar', '.gz']:
        with tarfile.open(filepath) as tar:
            tar.extractall(target_dir)
    else:
        raise ValueError('File format is not supported for unpacking.')


def archive_package(package_folder: 'Path') -> 'io.BytesIO':
    """
    Archives the given folder in zip format and return a data stream.
    :param package_folder: the folder path of the package
    :return: the data stream of zip content
    """

    with ImportExtensions(required=True):
        import pathspec

    root_path = package_folder.resolve()

    gitignore = root_path / '.gitignore'
    if not gitignore.exists():
        gitignore = Path(__resources_path__) / 'Python.gitignore'

    with gitignore.open() as fp:
        ignore_lines = [
            line.strip() for line in fp if line.strip() and (not line.startswith('#'))
        ]
        ignore_lines += ['.git', '.jina']
        ignored_spec = pathspec.PathSpec.from_lines('gitwildmatch', ignore_lines)

    zip_stream = io.BytesIO()
    try:
        zfile = zipfile.ZipFile(zip_stream, 'w', compression=zipfile.ZIP_DEFLATED)
    except EnvironmentError as e:
        raise e

    def _zip(base_path, path, archive):

        for p in path.iterdir():
            rel_path = p.relative_to(base_path)
            if ignored_spec.match_file(rel_path):
                continue
            if p.is_dir():
                _zip(base_path, p, archive)
            else:
                archive.write(p, rel_path)

    _zip(root_path, root_path, zfile)

    zfile.close()
    zip_stream.seek(0)

    return zip_stream


def download_with_resume(
    url: str,
    target_dir: 'Path',
    filename: Optional[str] = None,
    md5sum: Optional[str] = None,
) -> 'Path':
    """
    Download file from url to target_dir, and check md5sum.
    Performs a HTTP(S) download that can be restarted if prematurely terminated.
    The HTTP server must support byte ranges.

    :param url: the URL to download
    :param target_dir: the target path for the file
    :param filename: the filename of the downloaded file
    :param md5sum: the MD5 checksum to match

    :return: the filepath of the downloaded file
    """
    with ImportExtensions(required=True):
        import requests

    def _download(url, target, resume_byte_pos: int = None):
        resume_header = (
            {'Range': f'bytes={resume_byte_pos}-'} if resume_byte_pos else None
        )

        try:
            r = requests.get(url, stream=True, headers=resume_header)
        except requests.exceptions.RequestException as e:
            raise e

        block_size = 1024
        mode = 'ab' if resume_byte_pos else 'wb'

        with target.open(mode=mode) as f:
            for chunk in r.iter_content(32 * block_size):
                f.write(chunk)

    if filename is None:
        filename = url.split('/')[-1]
    filepath = target_dir / filename

    head_info = requests.head(url)
    file_size_online = int(head_info.headers.get('content-length', 0))

    _resume_byte_pos = None
    if filepath.exists():
        if md5sum and md5file(filepath) == md5sum:
            return filepath

        file_size_offline = filepath.stat().st_size
        if file_size_online > file_size_offline:
            _resume_byte_pos = file_size_offline

    _download(url, filepath, _resume_byte_pos)

    if md5sum and not md5file(filepath) == md5sum:
        raise RuntimeError('MD5 checksum failed.')

    return filepath


def upload_file(
    url: str,
    file_name: str,
    buffer_data: bytes,
    dict_data: Dict,
    headers: Dict,
    stream: bool = False,
    method: str = 'post',
):
    """Upload file to target url

    :param url: target url
    :param file_name: the file name
    :param buffer_data: the data to upload
    :param dict_data: the dict-style data to upload
    :param headers: the request header
    :param stream: receive stream response
    :param method: the request method
    :return: the response of request
    """
    with ImportExtensions(required=True):
        import requests

    dict_data.update({'file': (file_name, buffer_data)})

    (data, ctype) = requests.packages.urllib3.filepost.encode_multipart_formdata(
        dict_data
    )

    headers.update({'Content-Type': ctype})

    response = getattr(requests, method)(url, data=data, headers=headers, stream=stream)

    return response


def disk_cache_offline(
    cache_file: str = 'disk_cache.db',
    message: str = 'Calling {func_name} failed, using cached results',
):
    """
    Decorator which caches a function in disk and uses cache when a urllib.error.URLError exception is raised
    If the function was called with a kwarg force=True, then this decorator will always attempt to call it, otherwise,
    will always default to local cache.

    :param cache_file: the cache file
    :param message: the warning message shown when defaulting to cache. Use "{func_name}" if you want to print
        the function name

    :return: function decorator
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):

            call_hash = f'{func.__name__}({", ".join(map(str, args))})'

            pickle_protocol = 4
            file_lock = nullcontext()
            with ImportExtensions(
                required=False,
                help_text=f'FileLock is needed to guarantee non-concurrent access to the'
                f'cache_file {cache_file}',
            ):
                import filelock

                file_lock = filelock.FileLock(f'{cache_file}.lock', timeout=-1)

            cache_db = None
            with file_lock:
                try:
                    cache_db = shelve.open(
                        cache_file, protocol=pickle_protocol, writeback=True
                    )
                except Exception:
                    if os.path.exists(cache_file):
                        # cache is in an unsupported format, reset the cache
                        os.remove(cache_file)
                        cache_db = shelve.open(
                            cache_file, protocol=pickle_protocol, writeback=True
                        )

            if cache_db is None:
                # if we failed to load cache, do not raise, it is only an optimization thing
                return func(*args, **kwargs)
            else:
                with cache_db as dict_db:
                    try:
                        if call_hash in dict_db and not kwargs.get('force', False):
                            return dict_db[call_hash]

                        result = func(*args, **kwargs)
                        dict_db[call_hash] = result
                    except urllib.error.URLError:
                        if call_hash in dict_db:
                            default_logger.warning(
                                message.format(func_name=func.__name__)
                            )
                            return dict_db[call_hash]
                        else:
                            raise
                return result

        return wrapper

    return decorator


def is_requirements_installed(
    requirements_file: 'Path', show_warning: bool = False
) -> bool:
    """Return True if requirements.txt is installed locally
    :param requirements_file: the requirements.txt file
    :param show_warning: if to show a warning when a dependency is not satisfied
    :return: True or False if not satisfied
    """
    from pkg_resources import (
        DistributionNotFound,
        VersionConflict,
        RequirementParseError,
    )
    import pkg_resources

    install_reqs, install_options = _get_install_options(requirements_file)

    if len(install_reqs) == 0:
        return True

    try:
        pkg_resources.require('\n'.join(install_reqs))
    except (DistributionNotFound, VersionConflict, RequirementParseError) as ex:
        if show_warning:
            warnings.warn(repr(ex))
        return isinstance(ex, VersionConflict)
    return True


def _get_install_options(requirements_file: 'Path', excludes: Tuple[str] = ('jina',)):
    import pkg_resources

    with requirements_file.open() as requirements:
        install_options = []
        install_reqs = []
        for req in requirements:
            req = req.strip()
            if (not req) or req.startswith('#'):
                continue
            elif req.startswith('-'):
                install_options.extend(req.split(' '))
            else:
                for req_spec in pkg_resources.parse_requirements(req):
                    if (
                        req_spec.project_name not in excludes
                        or len(req_spec.extras) > 0
                    ):
                        install_reqs.append(req)

    return install_reqs, install_options


def install_requirements(requirements_file: 'Path', timeout: int = 1000):
    """Install modules included in requirments file
    :param requirements_file: the requirements.txt file
    :param timeout: the socket timeout (default = 1000s)
    """

    if is_requirements_installed(requirements_file):
        return

    install_reqs, install_options = _get_install_options(requirements_file)

    subprocess.check_call(
        [
            sys.executable,
            '-m',
            'pip',
            'install',
            '--compile',
            f'--default-timeout={timeout}',
        ]
        + install_reqs
        + install_options
    )



import base64
from typing import Optional

from ....helper import random_identity, download_mermaid_url


class PlotMixin:
    """Provide helper functions for :class:`Document` to plot and visualize itself. """

    @property
    def _mermaid_id(self):
        if not hasattr(self, '__mermaid_id'):
            self.__mermaid_id = random_identity()
        return self.__mermaid_id

    @_mermaid_id.setter
    def _mermaid_id(self, m_id):
        self.__mermaid_id = m_id

    def __mermaid_str__(self):
        results = []
        from google.protobuf.json_format import MessageToDict

        content = MessageToDict(self._pb_body, preserving_proto_field_name=True)

        _id = f'{self._mermaid_id[:3]}~Document~'

        for idx, c in enumerate(self.chunks):
            results.append(
                f'{_id} --> "{idx + 1}/{len(self.chunks)}" {c._mermaid_id[:3]}~Document~: chunks'
            )
            results.append(c.__mermaid_str__())

        for idx, c in enumerate(self.matches):
            results.append(
                f'{_id} ..> "{idx + 1}/{len(self.matches)}" {c._mermaid_id[:3]}~Document~: matches'
            )
            results.append(c.__mermaid_str__())
        if 'chunks' in content:
            content.pop('chunks')
        if 'matches' in content:
            content.pop('matches')
        if content:
            results.append(f'class {_id}{{')
            for k, v in content.items():
                if isinstance(v, (str, int, float, bytes)):
                    results.append(f'+{k} {str(v)[:10]}')
                else:
                    results.append(f'+{k}({type(getattr(self, k, v))})')
            results.append('}')

        return '\n'.join(results)

    def _mermaid_to_url(self, img_type: str) -> str:
        """
        Rendering the current flow as a url points to a SVG, it needs internet connection

        :param img_type: the type of image to be generated
        :return: the url pointing to a SVG
        """
        if img_type == 'jpg':
            img_type = 'img'

        mermaid_str = (
            """
                                                            %%{init: {'theme': 'base', 'themeVariables': { 'primaryColor': '#FFC666'}}}%%
                                                            classDiagram
        
                                                                    """
            + self.__mermaid_str__()
        )

        encoded_str = base64.b64encode(bytes(mermaid_str.strip(), 'utf-8')).decode(
            'utf-8'
        )

        return f'https://mermaid.ink/{img_type}/{encoded_str}'

    def _ipython_display_(self):
        """Displays the object in IPython as a side effect"""
        self.plot(inline_display=True)

    def plot(self, output: Optional[str] = None, inline_display: bool = False) -> None:
        """
        Visualize the Document recursively.

        :param output: a filename specifying the name of the image to be created,
                    the suffix svg/jpg determines the file type of the output image
        :param inline_display: show image directly inside the Jupyter Notebook
        """
        image_type = 'svg'
        if output and output.endswith('jpg'):
            image_type = 'jpg'

        url = self._mermaid_to_url(image_type)
        showed = False
        if inline_display:
            try:
                from IPython.display import Image, display

                display(Image(url=url))
                showed = True
            except:
                # no need to panic users
                pass

        if output:
            download_mermaid_url(url, output)
        elif not showed:
            from ....logging.predefined import default_logger

            default_logger.info(f'Document visualization: {url}')



import os
import urllib.request
import webbrowser
from pathlib import Path

from jina import Flow
from jina.importer import ImportExtensions
from jina.logging.predefined import default_logger
from jina.logging.profile import ProgressBar
from jina.parsers.helloworld import set_hw_chatbot_parser
from jina.types.document.generators import from_csv

if __name__ == '__main__':
    from my_executors import MyTransformer, MyIndexer
else:
    from .my_executors import MyTransformer, MyIndexer


def _get_flow(args):
    """Ensure the same flow is used in hello world example and system test."""
    return (
        Flow(cors=True)
        .add(uses=MyTransformer, replicas=args.replicas)
        .add(uses=MyIndexer, workspace=args.workdir)
    )


def hello_world(args):
    """
    Execute the chatbot example.

    :param args: arguments passed from CLI
    """
    Path(args.workdir).mkdir(parents=True, exist_ok=True)

    with ImportExtensions(
        required=True,
        help_text='this demo requires Pytorch and Transformers to be installed, '
        'if you haven\'t, please do `pip install jina[torch,transformers]`',
    ):
        import transformers, torch

        assert [torch, transformers]  #: prevent pycharm auto remove the above line

    targets = {
        'covid-csv': {
            'url': args.index_data_url,
            'filename': os.path.join(args.workdir, 'dataset.csv'),
        }
    }

    # download the data
    download_data(targets, args.download_proxy, task_name='download csv data')

    # now comes the real work
    # load index flow from a YAML file

    f = _get_flow(args)

    # index it!
    with f:
        f.index(
            from_csv(
                targets['covid-csv']['filename'], field_resolver={'question': 'text'}
            ),
            show_progress=True,
        )

        # switch to REST gateway at runtime
        f.protocol = 'http'
        f.port_expose = args.port_expose

        url_html_path = 'file://' + os.path.abspath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)), 'static/index.html'
            )
        )
        try:
            webbrowser.open(url_html_path, new=2)
        except:
            pass  # intentional pass, browser support isn't cross-platform
        finally:
            default_logger.info(
                f'You should see a demo page opened in your browser, '
                f'if not, you may open {url_html_path} manually'
            )

        if not args.unblock_query_flow:
            f.block()


def download_data(targets, download_proxy=None, task_name='download fashion-mnist'):
    """
    Download data.

    :param targets: target path for data.
    :param download_proxy: download proxy (e.g. 'http', 'https')
    :param task_name: name of the task
    """
    opener = urllib.request.build_opener()
    opener.addheaders = [('User-agent', 'Mozilla/5.0')]
    if download_proxy:
        proxy = urllib.request.ProxyHandler(
            {'http': download_proxy, 'https': download_proxy}
        )
        opener.add_handler(proxy)
    urllib.request.install_opener(opener)
    with ProgressBar(description=task_name) as t:
        for k, v in targets.items():
            if not os.path.exists(v['filename']):
                urllib.request.urlretrieve(
                    v['url'], v['filename'], reporthook=lambda *x: t.update(0.01)
                )


if __name__ == '__main__':
    args = set_hw_chatbot_parser().parse_args()
    hello_world(args)



import sys
from types import LambdaType
from typing import Callable, TYPE_CHECKING, Generator, Optional, overload, TypeVar

if TYPE_CHECKING:
    from ....helper import T, random_identity
    from ...document import Document
    from .... import DocumentArray

T_DA = TypeVar('T_DA')


class ParallelMixin:
    """Helper functions that provide parallel map to :class:`DocumentArray` or :class:`DocumentArrayMemmap`."""

    @overload
    def apply(
        self: 'T',
        func: Callable[['Document'], 'Document'],
        backend: str = 'process',
        num_worker: Optional[int] = None,
    ) -> 'T':
        """Apply each element in itself with ``func``, return itself after modified.

        :param func: a function that takes :class:`Document` as input and outputs :class:`Document`.
        :param backend: if to use multi-`process` or multi-`thread` as the parallelization backend. In general, if your
            ``func`` is IO-bound then perhaps `thread` is good enough. If your ``func`` is CPU-bound then you may use `process`.
            In practice, you should try yourselves to figure out the best value. However, if you wish to modify the elements
            in-place, regardless of IO/CPU-bound, you should always use `thread` backend.

            .. warning::
                When using `process` backend, you should not expect ``func`` modify elements in-place. This is because
                the multiprocessing backing pass the variable via pickle and work in another process. The passed object
                and the original object do **not** share the same memory.

        :param num_worker: the number of parallel workers. If not given, then the number of CPUs in the system will be used.

        """
        ...

    def apply(self: 'T', *args, **kwargs) -> 'T':
        """
        # noqa: DAR102
        # noqa: DAR101
        # noqa: DAR201
        :return: a new :class:`DocumentArray` or :class:`DocumentArrayMemmap`
        """
        new_da = type(self)()
        new_da.extend(self.map(*args, **kwargs))
        self.clear()
        self.extend(new_da)
        return self

    def map(
        self,
        func: Callable[['Document'], 'T'],
        backend: str = 'process',
        num_worker: Optional[int] = None,
    ) -> Generator['T', None, None]:
        """Return an iterator that applies function to every **element** of iterable in parallel, yielding the results.

        .. seealso::
            - To process on a batch of elements, please use :meth:`.map_batch`;
            - To return a :class:`DocumentArray`/:class:`DocumentArrayMemmap`, please use :meth:`.apply`.

        :param func: a function that takes :class:`Document` as input and outputs anything. You can either modify elements
            in-place (only with `thread` backend) or work later on return elements.
        :param backend: if to use multi-`process` or multi-`thread` as the parallelization backend. In general, if your
            ``func`` is IO-bound then perhaps `thread` is good enough. If your ``func`` is CPU-bound then you may use `process`.
            In practice, you should try yourselves to figure out the best value. However, if you wish to modify the elements
            in-place, regardless of IO/CPU-bound, you should always use `thread` backend.

            .. warning::
                When using `process` backend, you should not expect ``func`` modify elements in-place. This is because
                the multiprocessing backing pass the variable via pickle and work in another process. The passed object
                and the original object do **not** share the same memory.

        :param num_worker: the number of parallel workers. If not given, then the number of CPUs in the system will be used.
        :yield: anything return from ``func``
        """
        if _is_lambda_or_local_function(func) and backend == 'process':
            func = _globalize_lambda_function(func)
        with _get_pool(backend, num_worker) as p:
            for x in p.imap(func, self):
                yield x

    @overload
    def apply_batch(
        self: 'T',
        func: Callable[['DocumentArray'], 'DocumentArray'],
        batch_size: int,
        backend: str = 'process',
        num_worker: Optional[int] = None,
        shuffle: bool = False,
    ) -> 'T':
        """Apply each element in itself with ``func``, return itself after modified.

        :param func: a function that takes :class:`Document` as input and outputs :class:`Document`.
        :param backend: if to use multi-`process` or multi-`thread` as the parallelization backend. In general, if your
            ``func`` is IO-bound then perhaps `thread` is good enough. If your ``func`` is CPU-bound then you may use `process`.
            In practice, you should try yourselves to figure out the best value. However, if you wish to modify the elements
            in-place, regardless of IO/CPU-bound, you should always use `thread` backend.

            .. warning::
                When using `process` backend, you should not expect ``func`` modify elements in-place. This is because
                the multiprocessing backing pass the variable via pickle and work in another process. The passed object
                and the original object do **not** share the same memory.

        :param num_worker: the number of parallel workers. If not given, then the number of CPUs in the system will be used.
        :param batch_size: Size of each generated batch (except the last one, which might be smaller, default: 32)
        :param shuffle: If set, shuffle the Documents before dividing into minibatches.
        """
        ...

    def apply_batch(self: 'T', *args, **kwargs) -> 'T':
        """
        # noqa: DAR102
        # noqa: DAR101
        # noqa: DAR201
        :return: a new :class:`DocumentArray` or :class:`DocumentArrayMemmap`
        """
        new_da = type(self)()
        for _b in self.map_batch(*args, **kwargs):
            new_da.extend(_b)
        self.clear()
        self.extend(new_da)
        return self

    def map_batch(
        self: 'T_DA',
        func: Callable[['DocumentArray'], 'T'],
        batch_size: int,
        backend: str = 'process',
        num_worker: Optional[int] = None,
        shuffle: bool = False,
    ) -> Generator['T', None, None]:
        """Return an iterator that applies function to every **minibatch** of iterable in parallel, yielding the results.
        Each element in the returned iterator is :class:`DocumentArray`.

        .. seealso::
            - To process single element, please use :meth:`.map`;
            - To return :class:`DocumentArray` or :class:`DocumentArrayMemmap`, please use :meth:`.apply_batch`.

        :param batch_size: Size of each generated batch (except the last one, which might be smaller, default: 32)
        :param shuffle: If set, shuffle the Documents before dividing into minibatches.
        :param func: a function that takes :class:`DocumentArray` as input and outputs anything. You can either modify elements
            in-place (only with `thread` backend) or work later on return elements.
        :param backend: if to use multi-`process` or multi-`thread` as the parallelization backend. In general, if your
            ``func`` is IO-bound then perhaps `thread` is good enough. If your ``func`` is CPU-bound then you may use `process`.
            In practice, you should try yourselves to figure out the best value. However, if you wish to modify the elements
            in-place, regardless of IO/CPU-bound, you should always use `thread` backend.

            .. warning::
                When using `process` backend, you should not expect ``func`` modify elements in-place. This is because
                the multiprocessing backing pass the variable via pickle and work in another process. The passed object
                and the original object do **not** share the same memory.

        :param num_worker: the number of parallel workers. If not given, then the number of CPUs in the system will be used.
        :yield: anything return from ``func``
        """

        if _is_lambda_or_local_function(func) and backend == 'process':
            func = _globalize_lambda_function(func)
        with _get_pool(backend, num_worker) as p:
            for x in p.imap(func, self.batch(batch_size=batch_size, shuffle=shuffle)):
                yield x


def _get_pool(backend, num_worker):
    if backend == 'thread':
        from multiprocessing.pool import ThreadPool as Pool

        return Pool(processes=num_worker)
    elif backend == 'process':
        from multiprocessing.pool import Pool

        return Pool(processes=num_worker)
    else:
        raise ValueError(
            f'`backend` must be either `process` or `thread`, receiving {backend}'
        )


def _is_lambda_or_local_function(func):
    return (isinstance(func, LambdaType) and func.__name__ == '<lambda>') or (
        '<locals>' in func.__qualname__
    )


def _globalize_lambda_function(func):
    def result(*args, **kwargs):
        return func(*args, **kwargs)

    from ....helper import random_identity

    result.__name__ = result.__qualname__ = random_identity()
    setattr(sys.modules[result.__module__], result.__name__, result)
    return result



from typing import Type

from ...document import Document
from ....helper import T


class EmptyMixin:
    """Helper functions for building arrays with empty Document."""

    @classmethod
    def empty(cls: Type[T], size: int = 0) -> T:
        """Create a :class:`DocumentArray` or :class:`DocumentArrayMemmap` object with :attr:`size` empty
        :class:`Document` objects.

        :param size: the number of empty Documents in this container
        :return: a :class:`DocumentArray` or :class:`DocumentArrayMemmap` object
        """
        r = cls()
        r.extend(Document() for _ in range(size))
        return r



import itertools
import mmap
import os
import shutil
import tempfile
import warnings
from collections import OrderedDict
from collections.abc import MutableSequence
from pathlib import Path
from typing import Union, Iterable, Iterator, Optional, TYPE_CHECKING, List

import numpy as np

from .bpm import BufferPoolManager
from ..mixins import AllMixins
from .... import __windows__

_HEADER_NONE_ENTRY = (-1, -1, -1)
_PAGE_SIZE = mmap.ALLOCATIONGRANULARITY

if TYPE_CHECKING:
    from ...document import Document
    from ..document import DocumentArray

__all__ = ['DocumentArrayMemmap']


class DocumentArrayMemmap(
    AllMixins,
    MutableSequence,
):
    """
    Create a memory-map to an :class:`DocumentArray` stored in binary files on disk.

    Memory-mapped files are used for accessing :class:`Document` of large :class:`DocumentArray` on disk,
    without reading the entire file into memory.

    The :class:`DocumentArrayMemmap` on-disk storage consists of two files:
        - `header.bin`: stores id, offset, length and boundary info of each Document in `body.bin`;
        - `body.bin`: stores Documents continuously

    When loading :class:`DocumentArrayMemmap`, it loads the content of `header.bin` into memory, while storing
    all `body.bin` data on disk. As `header.bin` is often much smaller than `body.bin`, memory is saved.

    :class:`DocumentArrayMemmap` also loads a portion of the documents in a memory buffer and keeps the memory documents
    synced with the disk. This helps ensure that modified documents are persisted to the disk.
    The memory buffer size is configured with parameter `buffer_pool_size` which represents the number of documents
    that the buffer can store.

    .. note::
            To make sure the documents you modify are persisted to disk, make sure that the number of referenced
            documents does not exceed the buffer pool size. Otherwise, they won't be referenced by the buffer pool and
            they will not be persisted.
            The best practice is to always reference documents using DAM.

    This class is designed to work similarly as :class:`DocumentArray` but differs in the following aspects:
        - you can set the attribute of elements in a :class:`DocumentArrayMemmap` but you need to make sure that you
        don't reference more documents than the buffer pool size
        - each document

    To convert between a :class:`DocumentArrayMemmap` and a :class:`DocumentArray`

    .. highlight:: python
    .. code-block:: python

        # convert from DocumentArrayMemmap to DocumentArray
        dam = DocumentArrayMemmap('./tmp')
        ...

        da = DocumentArray(dam)

        # convert from DocumentArray to DocumentArrayMemmap
        dam2 = DocumentArrayMemmap('./tmp')
        dam2.extend(da)
    """

    def __init__(
        self,
        path: Optional[str] = None,
        key_length: int = 36,
        buffer_pool_size: int = 1000,
    ):
        if path:
            Path(path).mkdir(parents=True, exist_ok=True)
        else:
            path = tempfile.mkdtemp()
        self._path = path
        self._header_path = os.path.join(path, 'header.bin')
        self._body_path = os.path.join(path, 'body.bin')
        self._key_length = key_length
        self._last_mmap = None
        self._load_header_body()
        self._buffer_pool = BufferPoolManager(pool_size=buffer_pool_size)

    def insert(self, index: int, doc: 'Document') -> None:
        """Insert `doc` at `index`.

        :param index: the offset index of the insertion.
        :param doc: the doc needs to be inserted.
        """
        # This however must be here as inheriting from MutableSequence requires this method
        # TODO(team): implement this function if necessary and have time.
        raise NotImplementedError

    def reload(self):
        """Reload header of this object from the disk.

        This function is useful when another thread/process modify the on-disk storage and
        the change has not been reflected in this :class:`DocumentArray` object.

        This function only reloads the header, not the body.
        """
        self._load_header_body()
        self._buffer_pool.clear()

    def _load_header_body(self, mode: str = 'a'):
        if hasattr(self, '_header'):
            self._header.close()
        if hasattr(self, '_body'):
            self._body.close()

        open(self._header_path, mode).close()
        open(self._body_path, mode).close()

        self._header = open(self._header_path, 'r+b')
        self._body = open(self._body_path, 'r+b')

        tmp = np.frombuffer(
            self._header.read(),
            dtype=[
                ('', (np.str_, self._key_length)),  # key_length x 4 bytes
                ('', np.int64),  # 8 bytes
                ('', np.int64),  # 8 bytes
                ('', np.int64),  # 8 bytes
            ],
        )
        self._header_entry_size = 24 + 4 * self._key_length
        self._last_header_entry = len(tmp)

        self._header_map = OrderedDict()
        for idx, r in enumerate(tmp):
            if not np.array_equal((r[1], r[2], r[3]), _HEADER_NONE_ENTRY):
                self._header_map[r[0]] = (idx, r[1], r[2], r[3])

        self._header_keys = list(self._header_map.keys())

        self._body_fileno = self._body.fileno()
        self._start = 0
        if self._header_map:
            self._start = tmp[-1][1] + tmp[-1][3]
            self._body.seek(self._start)
        self._last_mmap = None

    def __len__(self):
        return len(self._header_map)

    def extend(self, docs: Iterable['Document']) -> None:
        """Extend the :class:`DocumentArrayMemmap` by appending all the items from the iterable.

        :param docs: the iterable of Documents to extend this array with
        """
        if not docs:
            return

        for d in docs:
            self.append(d, flush=False)
        self._header.flush()
        self._body.flush()
        self._last_mmap = None

    def clear(self) -> None:
        """Clear the on-disk data of :class:`DocumentArrayMemmap`"""
        self._load_header_body('wb')

    def _update_or_append(
        self,
        doc: 'Document',
        idx: Optional[int] = None,
        flush: bool = True,
        update_buffer: bool = True,
    ) -> None:
        value = bytes(doc)
        l = len(value)  #: the length
        p = int(self._start / _PAGE_SIZE) * _PAGE_SIZE  #: offset of the page
        r = (
            self._start % _PAGE_SIZE
        )  #: the remainder, i.e. the start position given the offset

        if idx is not None:
            self._header.seek(idx * self._header_entry_size, 0)

        if (doc.id is not None) and len(doc.id) > self._key_length:
            warnings.warn(
                f'The ID of doc ({doc.id}) will be truncated to the maximum length {self._key_length}'
            )

        self._header.write(
            np.array(
                (doc.id, p, r, r + l),
                dtype=[
                    ('', (np.str_, self._key_length)),
                    ('', np.int64),
                    ('', np.int64),
                    ('', np.int64),
                ],
            ).tobytes()
        )
        if idx is None:
            self._header_map[doc.id] = (self._last_header_entry, p, r, r + l)
            self._last_header_entry = self._last_header_entry + 1
            self._header_keys.append(doc.id)
        else:
            self._header_map[doc.id] = (idx, p, r, r + l)
            self._header_keys[idx] = doc.id
            self._header.seek(0, 2)
        self._start = p + r + l
        self._body.write(value)
        if flush:
            self._header.flush()
            self._body.flush()
            self._last_mmap = None
        if update_buffer:
            result = self._buffer_pool.add_or_update(doc.id, doc)
            if result:
                _key, _doc = result
                self._update(_doc, self._str2int_id(_key), update_buffer=False)

    def append(
        self, doc: 'Document', flush: bool = True, update_buffer: bool = True
    ) -> None:
        """
        Append `doc` in :class:`DocumentArrayMemmap`.

        :param doc: The doc needs to be appended.
        :param update_buffer: If set, update the buffer.
        :param flush: If set, then flush to disk on done.
        """
        self._update_or_append(doc, flush=flush, update_buffer=update_buffer)

    def _update(
        self, doc: 'Document', idx: int, flush: bool = True, update_buffer: bool = True
    ) -> None:
        """
        Update `doc` in :class:`DocumentArrayMemmap`.

        :param doc: The doc needed to be updated.
        :param idx: The position of the document.
        :param update_buffer: If set, update the buffer.
        :param flush: If set, then flush to disk on done.
        """
        self._update_or_append(doc, idx=idx, flush=flush, update_buffer=update_buffer)

    def _iteridx_by_slice(self, s: slice):
        length = self.__len__()
        start, stop, step = (
            s.start or 0,
            s.stop if s.stop is not None else length,
            s.step or 1,
        )
        if 0 > stop >= -length:
            stop = stop + length

        if 0 > start >= -length:
            start = start + length

        # if start and stop are in order, put them inside bounds
        # otherwise, range will return an empty iterator
        if start <= stop:
            if (start < 0 and stop < 0) or (start > length and stop > length):
                start, stop = 0, 0
            elif start < 0 and stop > length:
                start, stop = 0, length
            elif start < 0:
                start = 0
            elif stop > length:
                stop = length

        return range(start, stop, step)

    def _get_doc_array_by_slice(self, s: slice):
        from ..document import DocumentArray

        da = DocumentArray()
        for i in self._iteridx_by_slice(s):
            da.append(self[self._int2str_id(i)])

        return da

    @property
    def _mmap(self) -> 'mmap':
        if self._last_mmap is None:
            self._last_mmap = (
                mmap.mmap(self._body_fileno, length=0)
                if __windows__
                else mmap.mmap(self._body_fileno, length=0, prot=mmap.PROT_READ)
            )
        if __windows__:
            self._body.seek(self._start)
        return self._last_mmap

    def _get_doc_by_key(self, key: str):
        """
        returns a document by key (ID) from disk

        :param key: id of the document
        :return: returns a document
        """
        pos_info = self._header_map[key]
        _, p, r, r_plus_l = pos_info
        from ...document import Document

        return Document(self._mmap[p + r : p + r_plus_l])

    def __getitem__(self, key: Union[int, str, slice, List]):
        if isinstance(key, str):
            if key in self._buffer_pool:
                return self._buffer_pool[key]
            doc = self._get_doc_by_key(key)
            result = self._buffer_pool.add_or_update(key, doc)
            if result:
                _key, _doc = result
                self._update(_doc, self._str2int_id(_key), update_buffer=False)
            return doc

        elif isinstance(key, int):
            return self[self._int2str_id(key)]
        elif isinstance(key, slice):
            return self._get_doc_array_by_slice(key)
        elif isinstance(key, list):
            from ..document import DocumentArray

            return DocumentArray(self[k] for k in key)
        else:
            raise TypeError(f'`key` must be int, str or slice, but receiving {key!r}')

    def _del_doc(self, idx: int, str_key: str):
        p = idx * self._header_entry_size
        self._header.seek(p, 0)

        self._header.write(
            np.array(
                (str_key, -1, -1, -1),
                dtype=[
                    ('', (np.str_, self._key_length)),
                    ('', np.int64),
                    ('', np.int64),
                    ('', np.int64),
                ],
            ).tobytes()
        )
        self._header.seek(0, 2)
        self._header.flush()
        self._last_mmap = None
        pop_idx = self._header_keys.index(str_key)
        self._header_map.pop(str_key)
        self._header_keys.pop(pop_idx)
        self._buffer_pool.delete_if_exists(str_key)

    def __delitem__(self, key: Union[int, str, slice]):
        if isinstance(key, str):
            idx = self._str2int_id(key)
            str_key = key
            self._del_doc(idx, str_key)
        elif isinstance(key, int):
            idx = key
            str_key = self._int2str_id(idx)
            self._del_doc(idx, str_key)
        elif isinstance(key, slice):
            for idx in reversed(self._iteridx_by_slice(key)):
                str_key = self._int2str_id(idx)
                self._del_doc(idx, str_key)
        else:
            raise TypeError(f'`key` must be int, str or slice, but receiving {key!r}')

    def _str2int_id(self, key: str) -> int:
        return self._header_map[key][0]

    def _int2str_id(self, key: int) -> str:
        # i < 0 needs to be handled
        return self._header_keys[key]

    def __iter__(self) -> Iterator['Document']:
        for k in self._header_map.keys():
            yield self[k]

    def __setitem__(self, key: Union[int, str], value: 'Document') -> None:
        if isinstance(key, int):
            if 0 <= key < len(self):
                str_key = self._int2str_id(key)
                # override an existing entry
                self._update(value, key)

                # allows overwriting an existing document
                if str_key != value.id:
                    entry = self._header_map.pop(value.id)
                    self._header_map = OrderedDict(
                        [
                            (value.id, entry) if k == str_key else (k, v)
                            for k, v in self._header_map.items()
                        ]
                    )
                    if str_key in self._buffer_pool.doc_map:
                        self._buffer_pool.doc_map.pop(str_key)
            else:
                raise IndexError(f'`key`={key} is out of range')
        elif isinstance(key, str):
            if key != value.id:
                raise ValueError('key must be equal to document id')
            self._update(value, self._str2int_id(key))
        else:
            raise TypeError(f'`key` must be int or str, but receiving {key!r}')

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self._header_path == other._header_path
            and self._body_path == other._body_path
        )

    def __contains__(self, item: str):
        return item in self._header_map

    def flush(self) -> None:
        """Persists memory loaded documents to disk"""
        docs_to_flush = self._buffer_pool.docs_to_flush()
        for key, doc in docs_to_flush:
            self._update(doc, self._str2int_id(key), flush=False)
        self._header.flush()
        self._body.flush()
        self._last_mmap = None

    def __del__(self):
        try:
            self.flush()
        except:
            warnings.warn(f'{self!r} is not correctly flush to disk on deconstruction.')

    def prune(self) -> None:
        """Prune deleted Documents from this object, this yields a smaller on-disk storage. """
        dam = DocumentArrayMemmap(key_length=self._key_length)
        dam.extend(self)
        dam.reload()
        if hasattr(self, '_body'):
            self._body.close()
        os.remove(self._body_path)
        if hasattr(self, '_header'):
            self._header.close()
        os.remove(self._header_path)
        shutil.copy(os.path.join(dam.path, 'header.bin'), self._header_path)
        shutil.copy(os.path.join(dam.path, 'body.bin'), self._body_path)
        self.reload()

    @property
    def physical_size(self) -> int:
        """Return the on-disk physical size of this DocumentArrayMemmap, in bytes

        :return: the number of bytes
        """
        return os.stat(self._header_path).st_size + os.stat(self._body_path).st_size

    @staticmethod
    def _flatten(sequence):
        return itertools.chain.from_iterable(sequence)

    @property
    def path(self) -> str:
        """Get the path where the instance is stored.

        :returns: The stored path of the memmap instance.
        """
        return self._path

    @property
    def _pb_body(self):
        for v in self:
            yield v.proto



from typing import TYPE_CHECKING, TypeVar, Tuple, Sequence, Iterator, Optional

import numpy as np

from ..mixin import ProtoTypeMixin
from ...proto import jina_pb2

if TYPE_CHECKING:
    # fix type-hint complain for sphinx and flake
    import scipy.sparse
    import tensorflow
    import torch

    ArrayType = TypeVar(
        'ArrayType',
        np.ndarray,
        scipy.sparse.spmatrix,
        tensorflow.SparseTensor,
        tensorflow.Tensor,
        torch.Tensor,
        Sequence[float],
    )

    from ... import Document

__all__ = ['NdArray']


class NdArray(ProtoTypeMixin):
    """
    A base class for containing the protobuf message of NdArray. It defines interfaces for easier get/set value.

    Do not use this class directly. Subclass should be used.

    :param proto: the protobuf message, when not given then create a new one via :meth:`get_null_proto`
    """

    def __init__(self, proto: Optional[jina_pb2.NdArrayProto] = None):
        self._pb_body = proto if proto is not None else jina_pb2.NdArrayProto()

    def numpy(self) -> 'np.ndarray':
        """Return the value always in :class:`numpy.ndarray` regardless the framework type.

        :return: the value in :class:`numpy.ndarray`.
        """
        v = self.value
        if self.is_sparse:
            if hasattr(v, 'todense'):
                v = v.todense()
            elif hasattr(v, 'to_dense'):
                v = v.to_dense()
            elif self.framework == 'tensorflow':
                import tensorflow as tf

                if isinstance(v, tf.SparseTensor):
                    v = tf.sparse.to_dense(v)

        if hasattr(v, 'numpy'):
            v = v.numpy()
        return v

    @property
    def value(self) -> 'ArrayType':
        """Return the value in original framework type

        :return: the value of in numpy, scipy, tensorflow, pytorch type."""

        if self.is_sparse:
            if self.framework == 'scipy':
                idx, val, shape = self._get_raw_sparse_array()
                from scipy.sparse import coo_matrix

                x = coo_matrix((val, idx.T), shape=shape)
                sp_format = self._pb_body.parameters['sparse_format']
                if sp_format == 'bsr':
                    return x.tobsr()
                elif sp_format == 'csc':
                    return x.tocsc()
                elif sp_format == 'csr':
                    return x.tocsr()
                elif sp_format == 'coo':
                    return x
            elif self.framework == 'tensorflow':
                idx, val, shape = self._get_raw_sparse_array()
                from tensorflow import SparseTensor

                return SparseTensor(idx, val, shape)
            elif self.framework == 'torch':
                idx, val, shape = self._get_raw_sparse_array()
                from torch import sparse_coo_tensor

                return sparse_coo_tensor(idx, val, shape)
        else:
            if self.framework in {'numpy', 'torch', 'paddle', 'tensorflow'}:
                x = _get_dense_array(self._pb_body.dense)
                return _to_framework_array(x, self.framework)

    @staticmethod
    def ravel(value: 'ArrayType', docs: Iterator['Document'], field: str) -> None:
        """Ravel :attr:`value` into ``doc.field`` of each documents

        :param docs: the docs to set
        :param field: the field of the doc to set
        :param value: the value to be set on ``doc.field``
        """

        use_get_row = False
        if hasattr(value, 'getformat'):
            # for scipy only
            sp_format = value.getformat()
            if sp_format in {'bsr', 'coo'}:
                # for BSR and COO, they dont implement [j, ...] in scipy
                # but they offer get_row() API which implicitly translate the
                # sparse row into CSR format, hence needs to convert back
                # not very efficient, but this is the best we can do.
                use_get_row = True

        if use_get_row:
            emb_shape0 = value.shape[0]
            for d, j in zip(docs, range(emb_shape0)):
                row = getattr(value.getrow(j), f'to{sp_format}')()
                setattr(d, field, row)
        elif isinstance(value, (list, tuple)):
            for d, j in zip(docs, value):
                setattr(d, field, j)
        else:
            emb_shape0 = value.shape[0]
            for d, j in zip(docs, range(emb_shape0)):
                setattr(d, field, value[j, ...])

    @staticmethod
    def unravel(protos: Sequence[jina_pb2.NdArrayProto]) -> 'ArrayType':
        """Unravel many ndarray-like proto in one-shot, by following the shape
        and dtype of the first proto.

        :param protos: a list of ndarray protos
        :return: a framework ndarray
        """
        first = NdArray(protos[0])
        framework, is_sparse = first.framework, first.is_sparse

        if is_sparse:
            if framework in {'tensorflow'}:
                raise NotImplementedError(
                    f'fast ravel on sparse {framework} is not supported yet.'
                )

            val = _unravel_dense_array(
                (d.sparse.values.buffer for d in protos),
                shape=[],
                dtype=first.sparse.values.dtype,
            )

            shape = [len(protos)] + list(first.sparse.shape)

            all_ds = []
            for j, p in enumerate(protos):
                _d = _get_dense_array(p.sparse.indices)

                if framework == 'torch':
                    _idx = np.array([j] * _d.shape[-1], dtype=np.int32).reshape([1, -1])
                    _d = np.vstack([_idx, _d])
                if framework == 'scipy':
                    _idx = np.array([j] * _d.shape[0], dtype=np.int32)
                    _d = np.stack([_idx, _d[:, 1]], axis=-1)
                all_ds.append(_d)

            if framework == 'torch':
                idx = np.concatenate(all_ds, axis=-1)
                from torch import sparse_coo_tensor

                return sparse_coo_tensor(idx, val, shape)
            if framework == 'scipy':
                # scipy sparse is limited to ndim=2
                idx = np.concatenate(all_ds, axis=0)
                sp_format = first._pb_body.parameters['sparse_format']
                shape = [len(protos), first.sparse.shape[-1]]
                if sp_format == 'csc':
                    from scipy.sparse import csc_matrix

                    return csc_matrix((val, idx.T), shape=shape)
                if sp_format == 'csr':
                    from scipy.sparse import csr_matrix

                    return csr_matrix((val, idx.T), shape=shape)
                if sp_format == 'coo':
                    from scipy.sparse import coo_matrix

                    return coo_matrix((val, idx.T), shape=shape)
                if sp_format == 'bsr':
                    from scipy.sparse import bsr_matrix

                    return bsr_matrix((val, idx.T), shape=shape)

        else:
            if framework in {'numpy', 'torch', 'paddle', 'tensorflow'}:
                x = _unravel_dense_array(
                    (d.dense.buffer for d in protos),
                    shape=list(first.dense.shape),
                    dtype=first.dense.dtype,
                )
                return _to_framework_array(x, framework)

    @value.setter
    def value(self, value: 'ArrayType'):
        """Set the value from numpy, scipy, tensorflow, pytorch type to protobuf.

        :param value: the framework ndarray to be set.
        """
        framework, is_sparse = get_array_type(value)

        if framework == 'jina':
            # it is Jina's NdArray, simply copy it
            self._pb_body.cls_name = 'numpy'
            self._pb_body.CopyFrom(value._pb_body)
        elif framework == 'jina_proto':
            self._pb_body.cls_name = 'numpy'
            self._pb_body.CopyFrom(value)
        else:
            if is_sparse:
                if framework == 'scipy':
                    self._pb_body.parameters['sparse_format'] = value.getformat()
                    self._set_scipy_sparse(value)
                if framework == 'tensorflow':
                    self._set_tf_sparse(value)
                if framework == 'torch':
                    self._set_torch_sparse(value)
            else:
                if framework == 'numpy':
                    self._pb_body.cls_name = 'numpy'
                    _set_dense_array(value, self._pb_body.dense)
                if framework == 'python':
                    self._pb_body.cls_name = 'numpy'
                    _set_dense_array(np.array(value), self._pb_body.dense)
                if framework == 'tensorflow':
                    self._pb_body.cls_name = 'tensorflow'
                    _set_dense_array(value.numpy(), self._pb_body.dense)
                if framework == 'torch':
                    self._pb_body.cls_name = 'torch'
                    _set_dense_array(value.detach().cpu().numpy(), self._pb_body.dense)
                if framework == 'paddle':
                    self._pb_body.cls_name = 'paddle'
                    _set_dense_array(value.numpy(), self._pb_body.dense)

    @property
    def is_sparse(self) -> bool:
        """Check if the object represents a sparse ndarray.

        :return: True if the underlying ndarray is sparse
        """
        return self._pb_body.WhichOneof('content') == 'sparse'

    @property
    def framework(self) -> str:
        """Return the framework name of this ndarray object

        :return: the framework name
        """
        return self._pb_body.cls_name

    def _set_scipy_sparse(self, value: 'scipy.sparse.spmatrix'):
        v = value.tocoo(copy=True)
        indices = np.stack([v.row, v.col], axis=1)
        _set_dense_array(indices, self._pb_body.sparse.indices)
        _set_dense_array(v.data, self._pb_body.sparse.values)
        self._pb_body.sparse.ClearField('shape')
        self._pb_body.sparse.shape.extend(v.shape)
        self._pb_body.cls_name = 'scipy'

    def _set_tf_sparse(self, value: 'tensorflow.SparseTensor'):
        _set_dense_array(value.indices.numpy(), self._pb_body.sparse.indices)
        _set_dense_array(value.values.numpy(), self._pb_body.sparse.values)
        self._pb_body.sparse.ClearField('shape')
        self._pb_body.sparse.shape.extend(value.shape)
        self._pb_body.cls_name = 'tensorflow'

    def _set_torch_sparse(self, value):
        _set_dense_array(
            value.coalesce().indices().numpy(), self._pb_body.sparse.indices
        )
        _set_dense_array(value.coalesce().values().numpy(), self._pb_body.sparse.values)
        self._pb_body.sparse.ClearField('shape')
        self._pb_body.sparse.shape.extend(list(value.size()))
        self._pb_body.cls_name = 'torch'

    def _get_raw_sparse_array(self):
        idx = _get_dense_array(self._pb_body.sparse.indices)
        val = _get_dense_array(self._pb_body.sparse.values)
        shape = list(self._pb_body.sparse.shape)
        return idx, val, shape


def _get_dense_array(source):
    if source.buffer:
        x = np.frombuffer(source.buffer, dtype=source.dtype)
        return x.reshape(source.shape)
    elif len(source.shape) > 0:
        return np.zeros(source.shape)


def _set_dense_array(value, target):
    target.buffer = value.tobytes()
    target.ClearField('shape')
    target.shape.extend(list(value.shape))
    target.dtype = value.dtype.str


def get_array_type(array: 'ArrayType') -> Tuple[str, bool]:
    """Get the type of ndarray without importing the framework

    :param array: any array, scipy, numpy, tf, torch, etc.
    :return: a tuple where the first element represents the framework, the second represents if it is sparse array
    """
    module_tags = array.__class__.__module__.split('.')
    class_name = array.__class__.__name__

    if isinstance(array, (list, tuple)):
        return 'python', False

    if 'numpy' in module_tags:
        return 'numpy', False

    if 'jina' in module_tags:
        if class_name == 'NdArray':
            return 'jina', False  # sparse or not is irrelevant

    if 'jina_pb2' in module_tags:
        if class_name == 'NdArrayProto':
            return 'jina_proto', False  # sparse or not is irrelevant

    if 'tensorflow' in module_tags:
        if class_name == 'SparseTensor':
            return 'tensorflow', True
        if class_name == 'Tensor' or class_name == 'EagerTensor':
            return 'tensorflow', False

    if 'torch' in module_tags and class_name == 'Tensor':
        return 'torch', array.is_sparse

    if 'paddle' in module_tags and class_name == 'Tensor':
        # Paddle does not support sparse tensor on 11/8/2021
        # https://github.com/PaddlePaddle/Paddle/issues/36697
        return 'paddle', False

    if 'scipy' in module_tags and 'sparse' in module_tags:
        return 'scipy', True

    raise TypeError(f'can not determine the array type: {module_tags}.{class_name}')


def _to_framework_array(x, framework):
    if framework == 'numpy':
        return x
    elif framework == 'tensorflow':
        from tensorflow import convert_to_tensor

        return convert_to_tensor(x)
    elif framework == 'torch':
        from torch import from_numpy

        return from_numpy(x)
    elif framework == 'paddle':
        from paddle import to_tensor

        return to_tensor(x)


def _unravel_dense_array(source, shape, dtype):
    x_mat = b''.join(source)
    shape = [-1] + shape
    return np.frombuffer(x_mat, dtype=dtype).reshape(shape)


import google.protobuf.json_format
from google.protobuf.json_format import _Printer, _Parser


class _PrinterWithNdArraySupport(_Printer):
    def _MessageToJsonObject(self, message):
        message_descriptor = message.DESCRIPTOR
        full_name = message_descriptor.full_name
        if full_name == 'jina.NdArrayProto':
            return NdArray(message).numpy().tolist()
        else:
            return super()._MessageToJsonObject(message)


class _ParseWithNdArraySupport(_Parser):
    def ConvertMessage(self, value, message):
        message_descriptor = message.DESCRIPTOR
        full_name = message_descriptor.full_name
        if full_name == 'jina.NdArrayProto':
            na = NdArray()
            na.value = value
            message.CopyFrom(na._pb_body)
            return message
        else:
            return super().ConvertMessage(value, message)


google.protobuf.json_format._Printer = _PrinterWithNdArraySupport
google.protobuf.json_format._Parser = _ParseWithNdArraySupport



from typing import TYPE_CHECKING

from ....helper import typename

if TYPE_CHECKING:
    from ..types import DocumentArraySourceType


class MagicMixin:
    """Magic helpers for DA/DAM. """

    def __bool__(self):
        """To simulate ```l = []; if l: ...```

        :return: returns true if the length of the array is larger than 0
        """
        return len(self) > 0

    def __repr__(self):
        return f'<{typename(self)} (length={len(self)}) at {id(self)}>'

    def __add__(self, other: 'DocumentArraySourceType'):
        v = type(self)()
        for doc in self:
            v.append(doc)
        for doc in other:
            v.append(doc)
        return v

    def __iadd__(self, other: 'DocumentArraySourceType'):
        for doc in other:
            self.append(doc)
        return self

    def __getstate__(self):
        r = dict(serialized=bytes(self))
        if hasattr(self, '_ref_doc'):
            r['ref_doc'] = bytes(self._ref_doc)
        return r

    def __setstate__(self, state):
        from ..document import DocumentArray

        da = DocumentArray.load_binary(state['serialized'])
        if 'ref_doc' in state:
            from ...document import Document

            ref_doc = Document(state['ref_doc'])
            self.__init__(da, ref_doc)
        else:
            self.__init__()
            self.extend(da)



from typing import Iterator, AsyncIterator, Union

from ...helper import get_or_reuse_loop


class AsyncRequestsIterator:
    """Iterator to allow async iteration of blocking/non-blocking iterator from the Client"""

    def __init__(self, iterator: Union[Iterator, AsyncIterator]) -> None:
        """Async request iterator
        TODO (Deepankar): accept `num_req`

        :param iterator: request iterator
        """
        self.iterator = iterator

    def iterator__next__(self):
        """
        Executed inside a `ThreadPoolExecutor` via `loop.run_in_executor` to avoid following exception.
        "StopIteration interacts badly with generators and cannot be raised into a Future"

        :return: next request or None
        """
        try:
            return self.iterator.__next__()
        except StopIteration:
            return None

    def __aiter__(self):
        return self

    async def __anext__(self):
        if isinstance(self.iterator, Iterator):
            """
            An `Iterator` indicates "blocking" code, which might block all tasks in the event loop.
            Hence we iterate in the default executor provided by asyncio.
            """
            request = await get_or_reuse_loop().run_in_executor(
                None, self.iterator__next__
            )

            """
            `iterator.__next__` can be executed directly and that'd raise `StopIteration` in the executor,
            which raises the following exception while chaining states in futures.
            "StopIteration interacts badly with generators and cannot be raised into a Future"
            To avoid that, we handle the raise by a `return None`
            """
            if request is None:
                raise StopAsyncIteration
        elif isinstance(self.iterator, AsyncIterator):
            # we assume that `AsyncIterator` doesn't block the event loop
            request = await self.iterator.__anext__()

        return request



from typing import Optional, Dict, Any

from fastapi import Depends, APIRouter, HTTPException

from ... import Runtime400Exception
from ..dependencies import PodDepends
from ...models import DaemonID, ContainerItem, ContainerStoreStatus, PodModel
from ...stores import pod_store as store

router = APIRouter(prefix='/pods', tags=['pods'])


@router.get(
    path='', summary='Get all alive Pods\' status', response_model=ContainerStoreStatus
)
async def _get_items():
    return store.status


@router.get(path='/arguments', summary='Get all accepted arguments of a Pod')
async def _fetch_pod_params():
    return PodModel.schema()['properties']


@router.post(
    path='',
    summary='Create a Pod',
    description='Create a Pod and add it to the store',
    status_code=201,
    response_model=DaemonID,
)
async def _create(pod: PodDepends = Depends(PodDepends)):
    try:
        return await store.add(
            id=pod.id,
            workspace_id=pod.workspace_id,
            params=pod.params,
            ports=pod.ports,
            envs=pod.envs,
            device_requests=pod.device_requests,
        )
    except Exception as ex:
        raise Runtime400Exception from ex


@router.put(
    path='/rolling_update/{id}',
    summary='Trigger a rolling_update operation on the Pod object',
)
async def _rolling_update(
    id: DaemonID,
    dump_path: Optional[str] = None,
    uses_with: Optional[Dict[str, Any]] = None,
):
    try:
        if dump_path is not None:
            if uses_with is not None:
                uses_with['dump_path'] = dump_path
            else:
                uses_with = {'dump_path': dump_path}

        return await store.rolling_update(id=id, uses_with=uses_with)
    except Exception as ex:
        raise Runtime400Exception from ex


@router.put(
    path='/scale/{id}',
    summary='Trigger a scale operation on the Pod object',
)
async def _scale(id: DaemonID, replicas: int):
    try:
        return await store.scale(id=id, replicas=replicas)
    except Exception as ex:
        raise Runtime400Exception from ex


@router.delete(
    path='',
    summary='Terminate all running Pods',
)
async def _clear_all():
    await store.clear()


@router.delete(
    path='/{id}',
    summary='Terminate a running Pod',
    description='Terminate a running Pod and release its resources',
)
async def _delete(id: DaemonID, workspace: bool = False):
    try:
        await store.delete(id=id, workspace=workspace)
    except KeyError:
        raise HTTPException(status_code=404, detail=f'{id} not found in store')


@router.get(
    path='/{id}', summary='Get status of a running Pod', response_model=ContainerItem
)
async def _status(id: DaemonID):
    try:
        return store[id]
    except KeyError:
        raise HTTPException(status_code=404, detail=f'{id} not found in store')



from enum import Enum
from typing import Dict

from pydantic import BaseModel

from jina.enums import GatewayProtocolType
from .containers import ContainerItem, ContainerStoreStatus
from .custom import build_pydantic_model
from .id import DaemonID
from .workspaces import WorkspaceItem, WorkspaceStoreStatus

FlowModel = build_pydantic_model(model_name='FlowModel', module='flow')
PodModel = build_pydantic_model(model_name='PodModel', module='pod')
PeaModel = build_pydantic_model(model_name='PeaModel', module='pea')


GATEWAY_RUNTIME_DICT = {
    GatewayProtocolType.GRPC: 'GRPCRuntime',
    GatewayProtocolType.WEBSOCKET: 'WebSocketRuntime',
    GatewayProtocolType.HTTP: 'HTTPRuntime',
}


class DaemonStatus(BaseModel):
    """Pydantic model for DaemonStatus"""

    jina: Dict
    envs: Dict
    workspaces: WorkspaceStoreStatus
    peas: ContainerStoreStatus
    pods: ContainerStoreStatus
    flows: ContainerStoreStatus
    used_memory: str



import json
import os.path
from contextlib import nullcontext
from typing import Union, TextIO, TYPE_CHECKING, Type, List

if TYPE_CHECKING:
    from .....helper import T


class JsonIOMixin:
    """Save/load a array into a JSON file."""

    def save_json(self, file: Union[str, TextIO]) -> None:
        """Save array elements into a JSON file.

        Comparing to :meth:`save_binary`, it is human-readable but slower to save/load and the file size larger.

        :param file: File or filename to which the data is saved.
        """
        if hasattr(file, 'write'):
            file_ctx = nullcontext(file)
        else:
            file_ctx = open(file, 'w')

        with file_ctx as fp:
            for d in self:
                json.dump(d.to_dict(), fp)
                fp.write('\n')

    @classmethod
    def load_json(cls: Type['T'], file: Union[str, TextIO]) -> 'T':
        """Load array elements from a JSON file.

        :param file: File or filename or a JSON string to which the data is saved.

        :return: a DocumentArrayLike object
        """

        if hasattr(file, 'read'):
            file_ctx = nullcontext(file)
        elif os.path.exists(file):
            file_ctx = open(file)
        else:
            file_ctx = nullcontext(json.loads(file))

        from ....document import Document

        with file_ctx as fp:
            da = cls()
            da.extend(Document(v) for v in fp)
            return da

    def to_list(self) -> List:
        """Convert the object into a Python list.

        .. note::
            Array like object such as :class:`numpy.ndarray` will be converted to Python list.

        :return: a Python list
        """
        return [d.to_dict() for d in self]

    def to_json(self) -> str:
        """Convert the object into a JSON string. Can be loaded via :meth:`.load_json`.

        :return: a Python list
        """
        return json.dumps(self.to_list())



from typing import Union, TextIO, BinaryIO, TYPE_CHECKING, Type

if TYPE_CHECKING:
    from .....helper import T


class CommonIOMixin:
    """The common IO helper function for arrays. """

    def save(
        self, file: Union[str, TextIO, BinaryIO], file_format: str = 'json'
    ) -> None:
        """Save array elements into a JSON, a binary file or a CSV file.

        :param file: File or filename to which the data is saved.
        :param file_format: `json` or `binary` or `csv`. JSON and CSV files are human-readable,
            but binary format gives much smaller size and faster save/load speed. Note that, CSV file has very limited
            compatability, complex DocumentArray with nested structure can not be restored from a CSV file.
        """
        if file_format == 'json':
            self.save_json(file)
        elif file_format == 'binary':
            self.save_binary(file)
        elif file_format == 'csv':
            self.save_csv(file)
        else:
            raise ValueError('`format` must be one of [`json`, `binary`, `csv`]')

    @classmethod
    def load(
        cls: Type['T'], file: Union[str, TextIO, BinaryIO], file_format: str = 'json'
    ) -> 'T':
        """Load array elements from a JSON or a binary file, or a CSV file.

        :param file: File or filename to which the data is saved.
        :param file_format: `json` or `binary` or `csv`. JSON and CSV files are human-readable,
            but binary format gives much smaller size and faster save/load speed. CSV file has very limited compatability,
            complex DocumentArray with nested structure can not be restored from a CSV file.

        :return: the loaded DocumentArray object
        """
        if file_format == 'json':
            return cls.load_json(file)
        elif file_format == 'binary':
            return cls.load_binary(file)
        elif file_format == 'csv':
            return cls.load_csv(file)
        else:
            raise ValueError('`format` must be one of [`json`, `binary`, `csv`]')



import asyncio
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod

from ...types.request import Request
from ...importer import ImportExtensions

if TYPE_CHECKING:
    from ...types.request import Response
    from ...logging.logger import JinaLogger


class AioHttpClientlet(ABC):
    """aiohttp session manager"""

    def __init__(self, url: str, logger: 'JinaLogger') -> None:
        """HTTP Client to be used with the streamer

        :param url: url to send http/websocket request to
        :param logger: jina logger
        """
        self.url = url
        self.logger = logger
        self.msg_recv = 0
        self.msg_sent = 0
        self.session = None

    @abstractmethod
    async def send_message(self):
        """Send message to Gateway"""
        ...

    @abstractmethod
    async def recv_message(self):
        """Receive message from Gateway"""
        ...

    async def __aenter__(self):
        """enter async context

        :return: start self
        """
        return await self.start()

    async def start(self):
        """Create ClientSession and enter context

        :return: self
        """
        with ImportExtensions(required=True):
            import aiohttp

        self.session = aiohttp.ClientSession()
        await self.session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close(exc_type, exc_val, exc_tb)

    async def close(self, *args, **kwargs):
        """Close ClientSession

        :param args: positional args
        :param kwargs: keyword args"""
        await self.session.__aexit__(*args, **kwargs)


class HTTPClientlet(AioHttpClientlet):
    """HTTP Client to be used with the streamer"""

    async def send_message(self, request: 'Request'):
        """Sends a POST request to the server

        :param request: request as dict
        :return: send post message
        """
        req_dict = request.to_dict()
        req_dict['exec_endpoint'] = req_dict['header']['exec_endpoint']
        req_dict['data'] = req_dict['data'].get('docs', None)
        return await self.session.post(url=self.url, json=req_dict).__aenter__()

    async def recv_message(self):
        """Receive message for HTTP (sleep)

        :return: await sleep
        """
        return await asyncio.sleep(1e10)


class WebsocketClientlet(AioHttpClientlet):
    """Websocket Client to be used with the streamer"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.websocket = None

    async def send_message(self, request: 'Request'):
        """Send request in bytes to the server.

        :param request: request object
        :return: send request as bytes awaitable
        """
        try:
            return await self.websocket.send_bytes(request.SerializeToString())
        except ConnectionResetError:
            self.logger.critical(f'server connection closed already!')

    async def send_eoi(self):
        """To confirm end of iteration, we send `bytes(True)` to the server.

        :return: send `bytes(True)` awaitable
        """
        try:
            return await self.websocket.send_bytes(bytes(True))
        except ConnectionResetError:
            # server might be in a `CLOSING` state while sending EOI signal
            # which raises a `ConnectionResetError`, this can be ignored.
            pass

    async def recv_message(self) -> 'Response':
        """Receive messages in bytes from server and convert to `Response`

        ..note::
            aiohttp allows only one task which can `receive` concurrently.
            we need to make sure we don't create multiple tasks with `recv_message`

        :yield: response objects received from server
        """
        async for response in self.websocket:
            response_bytes = response.data
            resp = Request(response_bytes)
            yield resp.as_typed_request(resp.request_type).as_response()

    async def __aenter__(self):
        await super().__aenter__()
        self.websocket = await self.session.ws_connect(url=self.url).__aenter__()
        return self



import json
import uuid
import mimetypes
from typing import (
    Dict,
    Iterable,
    Optional,
    TypeVar,
    Union,
    overload,
    TYPE_CHECKING,
    Tuple,
    Sequence,
)

import numpy as np
from google.protobuf import json_format

from .mixins import AllMixins
from .mixins.version import versioned
from ..mixin import ProtoTypeMixin
from ..ndarray import NdArray
from ..score import NamedScore
from ..score.map import NamedScoreMapping
from ..struct import StructView
from ...excepts import BadDocType
from ...helper import typename
from ...proto import jina_pb2

if TYPE_CHECKING:
    from ..arrays.chunk import ChunkArray
    from ..arrays.match import MatchArray
    from ..ndarray import ArrayType

    DocumentSourceType = TypeVar(
        'DocumentSourceType', jina_pb2.DocumentProto, bytes, str, Dict
    )

__all__ = ['Document']

_all_mime_types = set(mimetypes.types_map.values())
_all_doc_content_keys = {'content', 'blob', 'text', 'buffer', 'graph'}


class Document(AllMixins, ProtoTypeMixin):
    """
    :class:`Document` is one of the **primitive data type** in Jina.

    It offers a Pythonic interface to allow users access and manipulate
    :class:`jina.jina_pb2.DocumentProto` object without working with Protobuf itself.

    To create a :class:`Document` object, simply:

        .. highlight:: python
        .. code-block:: python

            from jina import Document
            d = Document()
            d.text = 'abc'

    Jina requires each Document to have a string id. You can set a custom one,
    or if non has been set a random one will be assigned.

    To access and modify the content of the document, you can use :attr:`text`, :attr:`blob`, and :attr:`buffer`.
    Each property is implemented with proper setter, to improve the integrity and user experience. For example,
    assigning ``doc.blob`` or ``doc.embedding`` can be simply done via:

        .. highlight:: python
        .. code-block:: python

            import numpy as np

            # to set as content
            d.content = np.random.random([10, 5])

            # to set as embedding
            d.embedding = np.random.random([10, 5])

    MIME type is auto set/guessed when setting :attr:`content` and :attr:`uri`

    :class:`Document` also provides multiple way to build from existing Document. You can build :class:`Document`
    from ``jina_pb2.DocumentProto``, ``bytes``, ``str``, and ``Dict``. You can also use it as view (i.e.
    weak reference when building from an existing ``jina_pb2.DocumentProto``). For example,

        .. highlight:: python
        .. code-block:: python

            a = DocumentProto()
            b = Document(a, copy=False)
            a.text = 'hello'
            assert b.text == 'hello'

    You can leverage the :meth:`convert_a_to_b` interface to convert between content forms.

    """

    # overload_inject_start_document
    @overload
    def __init__(
        self,
        *,
        adjacency: Optional[int] = None,
        blob: Optional['ArrayType'] = None,
        buffer: Optional[bytes] = None,
        chunks: Optional[Iterable['Document']] = None,
        embedding: Optional['ArrayType'] = None,
        granularity: Optional[int] = None,
        id: Optional[str] = None,
        location: Optional[Sequence[float]] = None,
        matches: Optional[Iterable['Document']] = None,
        mime_type: Optional[str] = None,
        modality: Optional[str] = None,
        offset: Optional[float] = None,
        parent_id: Optional[str] = None,
        tags: Optional[Union[Dict, StructView]] = None,
        text: Optional[str] = None,
        uri: Optional[str] = None,
        weight: Optional[float] = None,
        **kwargs,
    ):
        """
        :param adjacency: the adjacency of this Document
        :param blob: the blob content of thi Document
        :param buffer: the buffer bytes from this document
        :param chunks: the array of chunks of this document
        :param embedding: the embedding of this Document
        :param granularity: the granularity of this Document
        :param id: the id of this Document
        :param location: location info in a tuple.
        :param matches: the array of matches attached to this document
        :param mime_type: the mime_type of this Document
        :param modality: the modality of the document.
        :param offset: the offset
        :param parent_id: the parent id of this Document
        :param tags: a Python dict view of the tags.
        :param text: the text from this document content
        :param uri: the uri of this Document
        :param weight: the weight of the document
        :param kwargs: other parameters to be set _after_ the document is constructed
        """

    # overload_inject_end_document

    def __init__(
        self,
        document: Optional[Union['DocumentSourceType', 'Document']] = None,
        field_resolver: Dict[str, str] = None,
        copy: bool = False,
        **kwargs,
    ):
        """
        :param document: the document to construct from. If ``bytes`` is given
                then deserialize a :class:`DocumentProto`; ``dict`` is given then
                parse a :class:`DocumentProto` from it; ``str`` is given, then consider
                it as a JSON string and parse a :class:`DocumentProto` from it; finally,
                one can also give `DocumentProto` directly, then depending on the ``copy``,
                it builds a view or a copy from it.
        :param copy: when ``document`` is given as a :class:`DocumentProto` object, build a
                view (i.e. weak reference) from it or a deep copy from it.
        :param field_resolver: a map from field names defined in JSON, dict to the field
            names defined in Document.
        :param kwargs: other parameters to be set _after_ the document is constructed

        .. note::

            When ``document`` is a JSON string or Python dictionary object, the constructor will only map the values
            from known fields defined in Protobuf, all unknown fields are mapped to ``document.tags``. For example,

            .. highlight:: python
            .. code-block:: python

                d = Document({'id': '123', 'hello': 'world', 'tags': {'good': 'bye'}})

                assert d.id == '123'  # true
                assert d.tags['hello'] == 'world'  # true
                assert d.tags['good'] == 'bye'  # true
        """
        try:
            if isinstance(document, jina_pb2.DocumentProto):
                if copy:
                    self._pb_body = jina_pb2.DocumentProto()
                    self._pb_body.CopyFrom(document)
                else:
                    self._pb_body = document
            elif isinstance(document, bytes):
                self._pb_body = jina_pb2.DocumentProto()
                self._pb_body.ParseFromString(document)
            elif isinstance(document, (dict, str)):
                # note that not any dict can be parsed in this branch. As later we use Protobuf JSON parser, this
                # dict must be a valid JSON. The next lines make sure the dict is a valid json.
                if isinstance(document, dict):
                    try:
                        json.dumps(document)
                    except:
                        raise json_format.ParseError

                if isinstance(document, str):
                    document = json.loads(document)

                if field_resolver:
                    document = {
                        field_resolver.get(k, k): v for k, v in document.items()
                    }

                user_fields = set(document)
                support_fields = set(
                    self.attributes(
                        include_proto_fields_camelcase=True,
                        include_properties=False,
                    )
                )

                self._pb_body = jina_pb2.DocumentProto()
                if support_fields.issuperset(user_fields):
                    json_format.ParseDict(document, self._pb_body)
                else:
                    _intersect = support_fields.intersection(user_fields)
                    _remainder = user_fields.difference(_intersect)
                    if _intersect:
                        json_format.ParseDict(
                            {k: document[k] for k in _intersect}, self._pb_body
                        )
                    if _remainder:
                        support_prop = set(
                            self.attributes(
                                include_proto_fields=False, include_properties=True
                            )
                        )
                        _intersect2 = support_prop.intersection(_remainder)
                        _remainder2 = _remainder.difference(_intersect2)

                        if _intersect2:
                            self._set_attributes(
                                **{p: document[p] for p in _intersect2}
                            )

                        if _remainder2:
                            self._pb_body.tags.update(
                                {k: document[k] for k in _remainder}
                            )

            elif isinstance(document, Document):
                if copy:
                    self._pb_body = jina_pb2.DocumentProto()
                    self._pb_body.CopyFrom(document.proto)
                else:
                    self._pb_body = document.proto
            elif document is not None:
                # note ``None`` is not considered as a bad type
                raise ValueError(f'{typename(document)} is not recognizable')
            else:
                # create an empty document
                self._pb_body = jina_pb2.DocumentProto()
        except json_format.ParseError:
            # append everything to kwargs that is more powerful in setting attributes
            self._pb_body = jina_pb2.DocumentProto()
            if isinstance(document, dict):
                kwargs.update(document)
        except Exception as ex:
            raise BadDocType(f'Fail to construct Document from {document!r}') from ex

        if self._pb_body.id is None or not self._pb_body.id:
            self.id = uuid.uuid1().hex

        if kwargs:
            # check if there are mutually exclusive content fields
            if len(_all_doc_content_keys.intersection(kwargs.keys())) > 1:
                raise ValueError(
                    f'Document content fields are mutually exclusive, please provide only one of {_all_doc_content_keys}'
                )
            self._set_attributes(**kwargs)

    @property
    def weight(self) -> float:
        """
        :return: the weight of the document
        """
        return self._pb_body.weight

    @weight.setter
    def weight(self, value: float):
        """
        Set the weight of the document.

        :param value: the float weight of the document.
        """
        if value is None:
            self.pop('weight')
            return
        self._pb_body.weight = value

    @property
    def modality(self) -> str:
        """
        :return: the modality of the document."""
        return self._pb_body.modality

    @modality.setter
    def modality(self, value: str):
        """Set the modality of the document.

        :param value: The modality of the document
        """
        if value is None:
            self.pop('modality')
            return
        self._pb_body.modality = value

    @property
    def tags(self) -> StructView:
        """Return the `tags` field of this Document as a Python dict

        :return: a Python dict view of the tags.
        """
        return StructView(self._pb_body.tags)

    @tags.setter
    def tags(self, value: Union[Dict, StructView]):
        """Set the `tags` field of this Document to a Python dict

        :param value: a Python dict or a StructView
        """
        if value is None:
            self.pop('tags')
            return

        if isinstance(value, StructView):
            self._pb_body.tags.Clear()
            self._pb_body.tags.update(value._pb_body)
        elif isinstance(value, dict):
            self._pb_body.tags.Clear()
            self._pb_body.tags.update(value)
        else:
            raise TypeError(f'{typename(value)} {value!r} is not supported.')

    @property
    def id(self) -> str:
        """The document id in string.

        :return: the id of this Document
        """
        return self._pb_body.id

    @property
    def parent_id(self) -> str:
        """The document's parent id in string.

        :return: the parent id of this Document
        """
        return self._pb_body.parent_id

    @id.setter
    def id(self, value: str):
        """Set document id to a string value.

        :param value: id as string
        """
        if value is None:
            self.pop('id')
            return
        self._pb_body.id = str(value)

    @parent_id.setter
    def parent_id(self, value: str):
        """Set document's parent id to a string value.

        :param value: id as string
        """
        if value is None:
            self.pop('parent_id')
            return
        self._pb_body.parent_id = str(value)

    @property
    def blob(self) -> 'ArrayType':
        """Return ``blob``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

            This property will return the `blob` of the `Document` as a `Dense` or `Sparse` array depending on the actual
            proto instance stored. In the case where the `blob` stored is sparse, it will return them as a `coo` matrix.

        :return: the blob content of thi Document
        """
        return NdArray(self._pb_body.blob).value

    @blob.setter
    def blob(self, value: 'ArrayType'):
        """Set the `blob` to `value`.

        :param value: the array value to set the blob
        """
        if value is None:
            self.pop('blob')
            return
        NdArray(self._pb_body.blob).value = value

    @property
    def embedding(self) -> 'ArrayType':
        """Return ``embedding`` of the content of a Document.

         .. note::
            This property will return the `embedding` of the `Document` as a `Dense` or `Sparse` array depending on the actual
            proto instance stored. In the case where the `embedding` stored is sparse, it will return them as a `coo` matrix.

        :return: the embedding of this Document
        """
        return NdArray(self._pb_body.embedding).value

    @embedding.setter
    def embedding(self, value: 'ArrayType'):
        """Set the ``embedding`` of the content of a Document.

        :param value: the array value to set the embedding
        """
        if value is None:
            self.pop('embedding')
            return
        NdArray(self._pb_body.embedding).value = value

    @property
    @versioned
    def matches(self) -> 'MatchArray':
        """Get all matches of the current document.

        :return: the array of matches attached to this document
        """
        # Problem with cyclic dependency
        from ..arrays.match import MatchArray

        return MatchArray(self._pb_body.matches, reference_doc=self)

    @matches.setter
    def matches(self, value: Iterable['Document']):
        """Get all chunks of the current document.

        :param value: value to set
        """
        if value is None:
            self.pop('matches')
            return
        self.pop('matches')
        self.matches.extend(value)

    @property
    @versioned
    def chunks(self) -> 'ChunkArray':
        """Get all chunks of the current document.

        :return: the array of chunks of this document
        """
        # Problem with cyclic dependency
        from ..arrays.chunk import ChunkArray

        return ChunkArray(self._pb_body.chunks, reference_doc=self)

    @chunks.setter
    def chunks(self, value: Iterable['Document']):
        """Get all chunks of the current document.

        :param value: the array of chunks of this document
        """
        if value is None:
            self.pop('chunks')
            return
        self.pop('chunks')
        self.chunks.extend(value)

    @property
    def buffer(self) -> bytes:
        """Return ``buffer``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

        :return: the buffer bytes from this document
        """
        return self._pb_body.buffer

    @buffer.setter
    def buffer(self, value: bytes):
        """Set the ``buffer`` to `value`.

        :param value: the bytes value to set the buffer
        """
        if value is None:
            self.pop('buffer')
            return
        self._pb_body.buffer = value

    @property
    def text(self) -> str:
        """Return ``text``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

        :return: the text from this document content
        """
        return self._pb_body.text

    @text.setter
    def text(self, value: str):
        """Set the `text` to `value`

        :param value: the text value to set as content
        """
        if value is None:
            self.pop('text')
            self.mime_type = None
            return
        self._pb_body.text = value
        self.mime_type = 'text/plain'

    @property
    def uri(self) -> str:
        """Return the URI of the document.

        :return: the uri of this Document
        """
        return self._pb_body.uri

    @uri.setter
    def uri(self, value: str):
        """Set the URI of the document.

        .. note::
            :attr:`mime_type` will be updated accordingly

        :param value: acceptable URI/URL, raise ``ValueError`` when it is not a valid URI
        """
        if value is None:
            self.pop('uri')
            self.mime_type = None
            return
        self._pb_body.uri = value
        mime_type = mimetypes.guess_type(value)[0]
        if mime_type:
            self.mime_type = mime_type  # Remote http/https contents mime_type will not be recognized.

    @property
    def mime_type(self) -> str:
        """Get MIME type of the document

        :return: the mime_type of this Document
        """
        return self._pb_body.mime_type

    @mime_type.setter
    def mime_type(self, value: str):
        """Set MIME type of the document

        :param value: the acceptable MIME type, raise ``ValueError`` when MIME type is not
                recognizable.
        """
        if value is None:
            self.pop('mime_type')
            return
        if value in _all_mime_types:
            self._pb_body.mime_type = value
        elif value:
            # given but not recognizable, do best guess
            r = mimetypes.guess_type(f'*.{value}')[0]
            if r:
                self._pb_body.mime_type = r
            else:
                self._pb_body.mime_type = value

    @property
    def granularity(self) -> int:
        """Return the granularity of the document.

        :return: the granularity of this Document
        """
        return self._pb_body.granularity

    @granularity.setter
    def granularity(self, value: int):
        """Set the granularity of the document.

        :param value: the value of the granularity to be set
        """
        if value is None:
            self.pop('granularity')
            return
        self._pb_body.granularity = value

    @property
    def adjacency(self) -> int:
        """Return the adjacency of the document.

        :return: the adjacency of this Document
        """
        return self._pb_body.adjacency

    @adjacency.setter
    def adjacency(self, value: int):
        """Set the adjacency of the document.

        :param value: the value of the adjacency to be set
        """
        if value is None:
            self.pop('adjacency')
            return
        self._pb_body.adjacency = value

    @property
    def scores(self):
        """Return the scores of the document.

        :return: the scores attached to this document as `:class:NamedScoreMapping`
        """
        return NamedScoreMapping(self._pb_body.scores)

    @scores.setter
    def scores(
        self,
        value: Dict[
            str, Union[NamedScore, jina_pb2.NamedScoreProto, float, np.generic]
        ],
    ):
        """Sets the scores of the `Document`. Specially important to provide the ability to start `scores` as:

            .. highlight:: python
            .. code-block:: python

                from jina import Document
                from jina.types.score import NamedScore
                d = Document(scores={'euclidean': 5, 'cosine': NamedScore(value=0.5)})

        :param value: the dictionary to set the scores
        """
        if value is None:
            self.pop('scores')
            return
        scores = NamedScoreMapping(self._pb_body.scores)
        for k, v in value.items():
            scores[k] = v

    @property
    def evaluations(self) -> NamedScoreMapping:
        """Return the evaluations of the document.

        :return: the evaluations attached to this document as `:class:NamedScoreMapping`
        """
        return NamedScoreMapping(self._pb_body.evaluations)

    @evaluations.setter
    def evaluations(
        self,
        value: Dict[
            str, Union[NamedScore, jina_pb2.NamedScoreProto, float, np.generic]
        ],
    ):
        """Sets the evaluations of the `Document`. Specially important to provide the ability to start `evaluations` as:

            .. highlight:: python
            .. code-block:: python

                from jina import Document
                from jina.types.score import NamedScore
                d = Document(evaluations={'precision': 0.9, 'recall': NamedScore(value=0.5)})

        :param value: the dictionary to set the evaluations
        """
        if value is None:
            self.pop('evaluations')
            return
        scores = NamedScoreMapping(self._pb_body.evaluations)
        for k, v in value.items():
            scores[k] = v

    @property
    def location(self) -> Tuple[float]:
        """Get the location information.

        :return: location info in a tuple.
        """
        return tuple(self._pb_body.location)

    @location.setter
    def location(self, value: Sequence[float]):
        """Set the location information of this Document.

        Location could mean the start and end index of a string;
        could be x,y (top, left) coordinate of an image crop; could be timestamp of an audio clip.

        :param value: the location info to be set.
        """
        if value is None:
            self.pop('location')
            return
        self.pop('location')
        self._pb_body.location.extend(value)

    @property
    def offset(self) -> float:
        """Get the offset information of this Document.

        :return: the offset
        """
        return self._pb_body.offset

    @offset.setter
    def offset(self, value: float):
        """Set the offset of this Document

        :param value: the offset value to be set.
        """
        if value is None:
            self.pop('offset')

        self._pb_body.offset = value



from typing import TYPE_CHECKING, Dict, Optional

from jina.helper import colored
from .base import BaseStore
from .peas import PeaStore

if TYPE_CHECKING:
    from ..models import DaemonID


class PodStore(PeaStore):
    """A Store of Pods spawned as Containers by Daemon"""

    _kind = 'pod'

    @BaseStore.dump
    async def rolling_update(
        self, id: 'DaemonID', uses_with: Optional[Dict] = None
    ) -> 'DaemonID':
        """rolling_update the Pod inside the container

        :param id: id of the Pod
        :param uses_with: the uses_with arguments to update the executor in pod_name
        :raises KeyError: if id doesn't exist in the store
        :return: id of the Flow
        """

        if id not in self:
            raise KeyError(f'{colored(id, "red")} not found in store.')

        try:
            object = await self.PUT(
                url=f'{self[id].metadata.uri}/{self._kind}/rolling_update',
                params=None,
                json=uses_with,
            )
        except Exception as e:
            self._logger.error(f'Error while sending rolling_update to the Pod: \n{e}')
            raise
        else:
            self[id].arguments.object = object
            self._logger.success(f'{colored(id, "green")} is updated successfully')
            return id

    @BaseStore.dump
    async def scale(self, id: 'DaemonID', replicas: int) -> 'DaemonID':
        """Scale the Pod inside the container

        :param id: id of the Pod
        :param replicas: The number of replicas to scale to
        :raises KeyError: if id doesn't exist in the store
        :return: id of the Flow
        """

        if id not in self:
            raise KeyError(f'{colored(id, "red")} not found in store.')

        try:
            object = await self.PUT(
                url=f'{self[id].metadata.uri}/{self._kind}/scale',
                params={'replicas': replicas},
                json=None,
            )
        except Exception as e:
            self._logger.error(f'Error while scaling the Pod: \n{e}')
            raise
        else:
            self[id].arguments.object = object
            self._logger.success(f'{colored(id, "green")} is scaled successfully')
            return id
